# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import locale
import logging
from collections import defaultdict

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.stringprep import saslprep

from gajim.common import app
from gajim.common import passwords
from gajim.common import types
from gajim.common.const import ClientState
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.setting_values import AllSettingsT

from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.dialogs import ConfirmationDialog
from gajim.gtk.dialogs import DialogButton
from gajim.gtk.filechoosers import Filter
from gajim.gtk.menus import build_accounts_menu
from gajim.gtk.omemo_trust_manager import OMEMOTrustManager
from gajim.gtk.settings import DropDownSetting
from gajim.gtk.settings import SettingsBox
from gajim.gtk.settings import SettingsDialog
from gajim.gtk.structs import ExportHistoryParam
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import iterate_listbox_children
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import open_window
from gajim.gtk.widgets import GajimAppWindow

log = logging.getLogger("gajim.gtk.accounts")


class AccountsWindow(GajimAppWindow):
    def __init__(self) -> None:
        GajimAppWindow.__init__(
            self,
            name="AccountsWindow",
            title=_("Accounts"),
            default_width=700,
            default_height=550,
            add_window_padding=False,
        )

        self._need_relogin: dict[str, list[AllSettingsT]] = {}
        self._accounts: dict[str, Account] = {}

        self._menu = AccountMenu()
        self._settings = Settings()

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.append(self._menu)
        box.append(self._settings)
        self.set_child(box)

        for account in app.get_accounts_sorted():
            self.add_account(account, initial=True)

        self._connect(self._menu, "menu-activated", self._on_menu_activated)
        self._connect(self.window, "close-request", self._on_close_request)

    def _cleanup(self) -> None:
        pass

    def _on_menu_activated(
        self, _listbox: Gtk.ListBox, account: str, name: str
    ) -> None:
        if name == "back":
            self._settings.set_page("add-account")
        elif name == "remove":
            self.on_remove_account(account)
        else:
            self._settings.set_page(name)

    def update_account_label(self, account: str) -> None:
        self._accounts[account].update_account_label()

    def update_proxy_list(self) -> None:
        for account in self._accounts:
            self._settings.update_proxy_list(account)

    def _check_relogin(self) -> bool:
        for account, r_settings in self._need_relogin.items():
            settings = self._get_relogin_settings(account)
            active = app.settings.get_account_setting(account, "active")
            if settings != r_settings:
                self._need_relogin[account] = settings
                if active:
                    self._relog(account)
                    return True
                break

        return False

    def _relog(self, account: str) -> None:

        def relog():
            client = app.get_client(account)
            client.disconnect(gracefully=True, reconnect=True, destroy_client=True)
            self.close()

        ConfirmationDialog(
            _("Re-Login Now?"),
            _("To apply all changes instantly, you have to re-login."),
            [
                DialogButton.make("Cancel", text=_("_Later"), callback=self.close),
                DialogButton.make("Accept", text=_("_Re-Login"), callback=relog),
            ],
        )

    @staticmethod
    def _get_relogin_settings(account: str) -> list[AllSettingsT]:
        values: list[AllSettingsT] = []
        values.append(app.settings.get_account_setting(account, "client_cert"))
        values.append(app.settings.get_account_setting(account, "proxy"))
        values.append(app.settings.get_account_setting(account, "resource"))
        values.append(app.settings.get_account_setting(account, "use_custom_host"))
        values.append(app.settings.get_account_setting(account, "custom_host"))
        values.append(app.settings.get_account_setting(account, "custom_port"))
        return values

    @staticmethod
    def on_remove_account(account: str) -> None:
        open_window("RemoveAccount", account=account)

    def remove_account(self, account: str) -> None:
        del self._need_relogin[account]
        self._accounts[account].remove()

    def add_account(self, account: str, initial: bool = False) -> None:
        self._need_relogin[account] = self._get_relogin_settings(account)
        self._accounts[account] = Account(account, self._menu, self._settings)
        if not initial:
            self._accounts[account].show()

    def select_account(self, account: str, page: str | None = None) -> None:
        try:
            self._accounts[account].select(page)
        except KeyError:
            log.warning("select_account() failed, account %s not found", account)

    def enable_account(self, account: str, state: bool) -> None:
        self._accounts[account].enable_account(state)

    def _on_close_request(self, _widget: Gtk.ApplicationWindow) -> bool:
        if self._check_relogin():
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE


class Settings(Gtk.ScrolledWindow):
    def __init__(self):
        Gtk.ScrolledWindow.__init__(self)
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._stack = Gtk.Stack(vhomogeneous=False)
        self._stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self._stack.add_css_class("settings-stack")
        self._stack.add_named(AddNewAccountPage(), "add-account")

        self.set_child(self._stack)
        self._page_signal_ids: dict[GenericSettingPage, int] = {}
        self._pages: dict[str, list[GenericSettingPage]] = defaultdict(list)

    def add_page(self, page: GenericSettingPage) -> None:
        self._pages[page.account].append(page)
        self._stack.add_named(page, f"{page.account}-{page.name}")
        self._page_signal_ids[page] = page.connect_signal(self._stack)

    def set_page(self, name: str) -> None:
        self._stack.set_visible_child_name(name)

    def remove_account(self, account: str) -> None:
        for page in self._pages[account]:
            signal_id = self._page_signal_ids[page]
            del self._page_signal_ids[page]
            self._stack.disconnect(signal_id)
            self._stack.remove(page)
        del self._pages[account]

    def update_proxy_list(self, account: str) -> None:
        for page in self._pages[account]:
            if page.name != "connection":
                continue
            assert isinstance(page, ConnectionPage)
            page.update_proxy_entries()


class AccountMenu(Gtk.Box, SignalManager):

    __gsignals__ = {
        "menu-activated": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
    }

    def __init__(self) -> None:
        Gtk.Box.__init__(self, hexpand=False, width_request=160)
        SignalManager.__init__(self)

        self.add_css_class("accounts-menu")

        self._stack = Gtk.Stack()
        self._stack.set_transition_type(Gtk.StackTransitionType.SLIDE_LEFT)

        self._accounts_listbox = Gtk.ListBox()
        self._accounts_listbox.set_sort_func(self._sort_func)
        self._accounts_listbox.add_css_class("accounts-menu-listbox")
        self._connect(
            self._accounts_listbox, "row-activated", self._on_account_row_activated
        )

        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled.set_child(self._accounts_listbox)
        self._stack.add_named(scrolled, "accounts")
        self.append(self._stack)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    @staticmethod
    def _sort_func(row1: AccountRow, row2: AccountRow) -> int:
        return locale.strcoll(row1.label.lower(), row2.label.lower())

    def add_account(self, row: AccountRow) -> None:
        self._accounts_listbox.append(row)
        sub_menu = AccountSubMenu(row.account)
        self._stack.add_named(sub_menu, f"{row.account}-menu")

        self._connect(sub_menu, "row-activated", self._on_sub_menu_row_activated)

    def remove_account(self, row: AccountRow) -> None:
        if self._stack.get_visible_child_name() != "accounts":
            # activate 'back' button
            listbox = cast(Gtk.ListBox, self._stack.get_visible_child())
            back_row = cast(Gtk.ListBoxRow, listbox.get_row_at_index(1))
            back_row.emit("activate")
        self._accounts_listbox.remove(row)
        sub_menu = self._stack.get_child_by_name(f"{row.account}-menu")
        assert sub_menu is not None
        self._stack.remove(sub_menu)

    def _on_account_row_activated(self, _listbox: Gtk.ListBox, row: AccountRow) -> None:
        self._stack.set_visible_child_name(f"{row.account}-menu")
        listbox = cast(Gtk.ListBox, self._stack.get_visible_child())
        listbox_row = cast(Gtk.ListBoxRow, listbox.get_row_at_index(2))
        listbox_row.emit("activate")

    def _on_sub_menu_row_activated(
        self, listbox: AccountSubMenu, row: MenuItem
    ) -> None:
        if row.name == "back":
            self._stack.set_visible_child_full(
                "accounts", Gtk.StackTransitionType.OVER_RIGHT
            )

        if row.name in ("back", "remove"):
            self.emit("menu-activated", listbox.account, row.name)

        else:
            self.emit(
                "menu-activated", listbox.account, f"{listbox.account}-{row.name}"
            )

    def set_page(self, account: str, page_name: str) -> None:
        sub_menu = cast(
            AccountSubMenu, self._stack.get_child_by_name(f"{account}-menu")
        )
        sub_menu.select_row_by_name(page_name)
        self.emit("menu-activated", account, f"{account}-{page_name}")

    def update_account_label(self, account: str) -> None:
        self._accounts_listbox.invalidate_sort()
        sub_menu = cast(
            AccountSubMenu, self._stack.get_child_by_name(f"{account}-menu")
        )
        sub_menu.update()


class AccountSubMenu(Gtk.ListBox):

    __gsignals__ = {"update": (GObject.SignalFlags.RUN_FIRST, None, (str,))}

    def __init__(self, account: str) -> None:
        Gtk.ListBox.__init__(self)
        self.set_vexpand(True)
        self.set_hexpand(True)
        self.add_css_class("accounts-menu-listbox")

        self._account = account

        self.append(AccountLabelMenuItem(self, self._account))
        self.append(BackMenuItem())
        self.append(PageMenuItem("general", _("General")))
        self.append(PageMenuItem("privacy", _("Privacy")))
        self.append(PageMenuItem("encryption-omemo", _("Encryption (OMEMO)")))
        self.append(PageMenuItem("connection", _("Connection")))
        self.append(PageMenuItem("advanced", _("Advanced")))
        self.append(RemoveMenuItem())

    @property
    def account(self) -> str:
        return self._account

    def select_row_by_name(self, row_name: str) -> None:
        for row in iterate_listbox_children(self):
            if not isinstance(row, PageMenuItem):
                continue
            if row.name == row_name:
                self.select_row(row)
                return

    def update(self) -> None:
        self.emit("update", self._account)


class MenuItem(Gtk.ListBoxRow, SignalManager):
    def __init__(self, name: str) -> None:
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)
        self._name = name
        self._box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        self._label = Gtk.Label()

        self.set_child(self._box)

    @property
    def name(self) -> str:
        return self._name

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)


class RemoveMenuItem(MenuItem):
    def __init__(self) -> None:
        super().__init__("remove")
        self._label.set_text(_("Remove"))
        image = Gtk.Image.new_from_icon_name("user-trash-symbolic")

        self.set_selectable(False)
        image.add_css_class("error-color")

        self._box.append(image)
        self._box.append(self._label)


class AccountLabelMenuItem(MenuItem):
    def __init__(self, parent: AccountSubMenu, account: str) -> None:
        super().__init__("account-label")
        self._update_account_label(parent, account)

        self.set_selectable(False)
        self.set_sensitive(False)
        self.set_activatable(False)

        image = Gtk.Image.new_from_icon_name("avatar-default-symbolic")
        image.add_css_class("insensitive-fg-color")

        self._label.add_css_class("accounts-label-row")
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._label.set_xalign(0)

        self._box.append(image)
        self._box.append(self._label)

        self._connect(parent, "update", self._update_account_label)

    def _update_account_label(self, _listbox: Gtk.ListBox, account: str) -> None:
        account_label = app.get_account_label(account)
        self._label.set_text(account_label)


class BackMenuItem(MenuItem):
    def __init__(self) -> None:
        super().__init__("back")
        self.set_selectable(False)

        self._label.set_text(_("Back"))

        image = Gtk.Image.new_from_icon_name("go-previous-symbolic")
        image.add_css_class("insensitive-fg-color")

        self._box.append(image)
        self._box.append(self._label)


class PageMenuItem(MenuItem):
    def __init__(self, name: str, label: str) -> None:
        super().__init__(name)

        if name == "general":
            icon = "avatar-default-symbolic"
        elif name == "privacy":
            icon = "feather-eye-symbolic"
        elif name == "encryption-omemo":
            icon = "channel-secure-symbolic"
        elif name == "connection":
            icon = "feather-globe-symbolic"
        elif name == "advanced":
            icon = "preferences-system-symbolic"
        else:
            icon = "dialog-error-symbolic"

        image = Gtk.Image.new_from_icon_name(icon)
        self._label.set_text(label)

        self._box.append(image)
        self._box.append(self._label)


class Account:
    def __init__(self, account: str, menu: AccountMenu, settings: Settings) -> None:
        self._account = account
        self._menu = menu
        self._settings = settings

        self._settings.add_page(GeneralPage(account))
        self._settings.add_page(PrivacyPage(account))
        self._settings.add_page(EncryptionOMEMOPage(account))
        self._settings.add_page(ConnectionPage(account))
        self._settings.add_page(AdvancedPage(account))

        self._account_row = AccountRow(account)
        self._menu.add_account(self._account_row)

    def select(self, page_name: str | None = None) -> None:
        self._account_row.emit("activate")
        if page_name is not None:
            self._menu.set_page(self._account, page_name)

    def show(self) -> None:
        self.select()

    def remove(self) -> None:
        self._menu.remove_account(self._account_row)
        self._settings.remove_account(self._account)

    def update_account_label(self) -> None:
        self._account_row.update_account_label()
        self._menu.update_account_label(self._account)

    def enable_account(self, state: bool) -> None:
        self._account_row.enable_account(state)

    @property
    def menu(self) -> AccountMenu:
        return self._menu

    @property
    def account(self) -> str:
        return self._account

    @property
    def settings(self) -> str:
        return self._account


class AccountRow(Gtk.ListBoxRow, SignalManager):
    def __init__(self, account: str) -> None:
        Gtk.ListBoxRow.__init__(self, selectable=False)
        SignalManager.__init__(self)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)

        self._account = account

        self._label = Gtk.Label(label=app.get_account_label(account))
        self._label.set_halign(Gtk.Align.START)
        self._label.set_hexpand(True)
        self._label.set_ellipsize(Pango.EllipsizeMode.END)
        self._label.set_xalign(0)
        self._label.set_width_chars(18)

        next_icon = Gtk.Image.new_from_icon_name("go-next-symbolic")
        next_icon.add_css_class("insensitive-fg-color")

        account_enabled = app.settings.get_account_setting(self._account, "active")
        self._switch = Gtk.Switch()
        self._switch.set_active(account_enabled)
        self._switch.set_vexpand(False)

        self._switch_state_label = Gtk.Label()
        self._switch_state_label.set_xalign(0)
        self._switch_state_label.set_valign(Gtk.Align.CENTER)
        label_width = max(len(p_("Switch", "On")), len(p_("Switch", "Off")))
        self._switch_state_label.set_width_chars(label_width)
        self._set_label(account_enabled)

        self._connect(self._switch, "state-set", self._on_enable_switch, self._account)

        box.append(self._switch)
        box.append(self._switch_state_label)
        box.append(Gtk.Separator())
        box.append(self._label)
        box.append(next_icon)
        self.set_child(box)

    @property
    def account(self) -> str:
        return self._account

    @property
    def label(self) -> str:
        return self._label.get_text()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def update_account_label(self) -> None:
        self._label.set_text(app.get_account_label(self._account))

    def enable_account(self, state: bool) -> None:
        self._switch.set_state(state)
        self._set_label(state)

    def _set_label(self, active: bool) -> None:
        text = p_("Switch", "On") if active else p_("Switch", "Off")
        self._switch_state_label.set_text(text)

    def _on_enable_switch(self, switch: Gtk.Switch, state: bool, account: str) -> int:

        def _disable() -> None:
            client = app.get_client(account)
            client.connect_signal("state-changed", self._on_state_changed)
            client.change_status("offline", "offline")
            switch.set_state(state)
            self._set_label(state)

        account_is_active = app.settings.get_account_setting(account, "active")
        if account_is_active == state:
            return Gdk.EVENT_PROPAGATE

        if account_is_active and not app.get_client(account).state.is_disconnected:
            # Connecting or connected
            window = get_app_window("AccountsWindow")
            assert window is not None

            account_label = app.get_account_label(account)
            ConfirmationDialog(
                _("Disable Account?"),
                _(
                    "Account %(name)s is still connected\n"
                    "All chat and group chat windows will be closed."
                )
                % {"name": account_label},
                [
                    DialogButton.make(
                        "Cancel", callback=lambda: switch.set_active(True)
                    ),
                    DialogButton.make(
                        "Remove", text=_("_Disable Account"), callback=_disable
                    ),
                ],
                transient_for=window.window,
            )
            return Gdk.EVENT_STOP

        if state:
            app.app.enable_account(account)
        else:
            app.app.disable_account(account)

        return Gdk.EVENT_PROPAGATE

    def _on_state_changed(
        self, client: types.Client, _signal_name: str, client_state: ClientState
    ) -> None:

        if client_state.is_disconnected:
            app.app.disable_account(client.account)


class AddNewAccountPage(Gtk.Box):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=18)
        self.set_vexpand(True)
        self.set_hexpand(True)
        self.set_margin_top(24)
        image = Gtk.Image.new_from_icon_name("gajim-symbolic")
        image.set_pixel_size(100)
        image.set_opacity(0.2)
        self.append(image)

        button = Gtk.Button(label=_("Add Account"))
        button.add_css_class("suggested-action")
        button.set_action_name("app.add-account")
        button.set_halign(Gtk.Align.CENTER)
        self.append(button)


class GenericSettingPage(Gtk.Box):

    name = ""

    def __init__(self, account: str, settings: list[Setting]) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=12)
        self.set_valign(Gtk.Align.START)
        self.set_vexpand(True)
        self.add_css_class("settings-page")
        self.account = account

        self.listbox = SettingsBox(account)
        self.listbox.add_css_class("border")
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.listbox.set_vexpand(False)
        self.listbox.set_valign(Gtk.Align.END)

        for setting in settings:
            self.listbox.add_setting(setting)
        self.listbox.update_states()

        self.append(self.listbox)

    def connect_signal(self, stack: Gtk.Stack) -> int:
        return stack.connect("notify::visible-child", self._on_visible_child_changed)

    def _on_visible_child_changed(self, stack: Gtk.Stack, _param: Any) -> None:
        if self == stack.get_visible_child():
            self.listbox.update_states()


class GeneralPage(GenericSettingPage):

    name = "general"

    def __init__(self, account: str) -> None:

        workspaces = self._get_workspaces()

        settings = [
            Setting(
                SettingKind.ENTRY,
                _("Label"),
                SettingType.ACCOUNT_CONFIG,
                "account_label",
                callback=self._on_account_name_change,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Default Workspace"),
                SettingType.ACCOUNT_CONFIG,
                "default_workspace",
                props={"data": workspaces},
                desc=_("Chats from this account will use this workspace by default"),
            ),
            Setting(
                SettingKind.COLOR,
                _("Color"),
                SettingType.ACCOUNT_CONFIG,
                "account_color",
                desc=_("Recognize your account by color"),
            ),
            Setting(
                SettingKind.LOGIN,
                _("Login"),
                SettingType.DIALOG,
                desc=_("Change your account’s password, etc."),
                bind="account::anonymous_auth",
                inverted=True,
                props={"dialog": LoginDialog},
            ),
            # Currently not supported by nbxmpp
            #
            # Setting(SettingKind.DIALOG,
            #         _('Client Certificate'),
            #         SettingType.DIALOG,
            #         props={'dialog': CertificateDialog}),
            Setting(
                SettingKind.SWITCH,
                _("Connect on startup"),
                SettingType.ACCOUNT_CONFIG,
                "autoconnect",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Global Status"),
                SettingType.ACCOUNT_CONFIG,
                "sync_with_global_status",
                desc=_("Synchronize the status of all accounts"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Remember Last Status"),
                SettingType.ACCOUNT_CONFIG,
                "restore_last_status",
                desc=_("Restore status and status message of your last session"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use file transfer proxies"),
                SettingType.ACCOUNT_CONFIG,
                "use_ft_proxies",
            ),
        ]
        GenericSettingPage.__init__(self, account, settings)

    @staticmethod
    def _get_workspaces() -> dict[str, str]:
        workspaces: dict[str, str] = {"": _("Disabled")}
        for workspace_id in app.settings.get_workspaces():
            name = app.settings.get_workspace_setting(workspace_id, "name")
            workspaces[workspace_id] = name
        return workspaces

    def _on_account_name_change(self, *args: Any) -> None:
        window = get_app_window("AccountsWindow")
        assert window is not None
        window.update_account_label(self.account)
        build_accounts_menu()


class PrivacyPage(GenericSettingPage):

    name = "privacy"

    def __init__(self, account: str) -> None:
        self._account = account
        self._client: types.Client | None = None
        if app.account_is_connected(account):
            self._client = app.get_client(account)

        history_max_age = {
            -1: _("Forever"),
            0: _("Until Gajim is Closed"),
            86400: _("1 Day"),
            604800: _("1 Week"),
            2629743: _("1 Month"),
            7889229: _("3 Months"),
            15778458: _("6 Months"),
            31556926: _("1 Year"),
        }

        chatstate_entries = {
            "disabled": _("Disabled"),
            "composing_only": _("Composing Only"),
            "all": _("All Chat States"),
        }

        encryption_entries = {
            "": _("Unencrypted"),
            "OMEMO": "OMEMO",
            "OpenPGP": "OpenPGP",
            "PGP": "PGP",
        }

        param = ExportHistoryParam(account=account, jid=None)

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Default Encryption"),
                SettingType.ACCOUNT_CONFIG,
                "encryption_default",
                desc=_(
                    "Encryption method to use "
                    "unless overridden on a per-contact basis"
                ),
                props={"data": encryption_entries},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Idle Time"),
                SettingType.ACCOUNT_CONFIG,
                "send_idle_time",
                callback=self._send_idle_time,
                desc=_("Disclose the time of your last activity"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Local System Time"),
                SettingType.ACCOUNT_CONFIG,
                "send_time_info",
                callback=self._send_time_info,
                desc=_("Disclose the local system time of the device Gajim runs on"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Operating System"),
                SettingType.ACCOUNT_CONFIG,
                "send_os_info",
                callback=self._send_os_info,
                desc=_(
                    "Disclose information about the "
                    "operating system you currently use"
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Media Playback"),
                SettingType.ACCOUNT_CONFIG,
                "publish_tune",
                callback=self._publish_tune,
                desc=_(
                    "Disclose information about media that is "
                    "currently being played on your system."
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Ignore Unknown Contacts"),
                SettingType.ACCOUNT_CONFIG,
                "ignore_unknown_contacts",
                desc=_("Ignore everything from contacts not in your contact list"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Send Message Receipts"),
                SettingType.ACCOUNT_CONFIG,
                "answer_receipts",
                desc=_("Tell your contacts if you received a message"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Send Chatstate"),
                SettingType.ACCOUNT_CONFIG,
                "send_chatstate_default",
                desc=_("Default for chats"),
                props={
                    "data": chatstate_entries,
                    "button-text": _("Reset"),
                    "button-tooltip": _("Reset all chats to the current default value"),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_send_chatstate,
                },
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Send Chatstate in Group Chats"),
                SettingType.ACCOUNT_CONFIG,
                "gc_send_chatstate_default",
                desc=_("Default for group chats"),
                props={
                    "data": chatstate_entries,
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_gc_send_chatstate,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Send Read Markers"),
                SettingType.VALUE,
                app.settings.get_account_setting(account, "send_marker_default"),
                callback=self._send_read_marker,
                desc=_("Default for chats and private group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _("Reset all chats to the current default value"),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_send_read_marker,
                },
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Keep Chat History"),
                SettingType.ACCOUNT_CONFIG,
                "chat_history_max_age",
                props={"data": history_max_age},
                desc=_("How long Gajim should keep your chat history"),
            ),
            Setting(
                SettingKind.ACTION,
                _("Export Chat History"),
                SettingType.ACTION,
                "app.export-history",
                props={"variant": param.to_variant()},
                desc=_("Export your chat history from Gajim"),
            ),
        ]
        GenericSettingPage.__init__(self, account, settings)

    @staticmethod
    def _reset_send_chatstate(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_contact_settings("send_chatstate", None)

    @staticmethod
    def _reset_gc_send_chatstate(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("send_chatstate", None)

    def _send_idle_time(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("LastActivity").set_enabled(state)

    def _send_time_info(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("EntityTime").set_enabled(state)

    def _send_os_info(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("SoftwareVersion").set_enabled(state)

    def _publish_tune(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("UserTune").set_enabled(state)

    def _send_read_marker(self, state: bool, _data: Any) -> None:
        app.settings.set_account_setting(self._account, "send_marker_default", state)
        app.settings.set_account_setting(
            self._account, "gc_send_marker_private_default", state
        )

    def _reset_send_read_marker(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_contact_settings("send_marker", None)
        app.settings.set_group_chat_settings("send_marker", None, context="private")


class EncryptionOMEMOPage(GenericSettingPage):

    name = "encryption-omemo"

    def __init__(self, account: str) -> None:
        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Blind Trust"),
                SettingType.ACCOUNT_CONFIG,
                "omemo_blind_trust",
                desc=_("Blindly trust new devices until you verify them"),
            )
        ]
        GenericSettingPage.__init__(self, account, settings)

        btbv_label = Gtk.Label()
        btbv_label.set_xalign(0)
        wiki_url = "https://dev.gajim.org/gajim/gajim/-/wikis/help/OMEMO"
        link_text = _("Read more about blind trust")
        markup = f'<a href="{wiki_url}">{link_text}</a>'
        btbv_label.set_markup(markup)
        self.prepend(btbv_label)

        heading = Gtk.Label(label=_("Trust Management"))
        heading.add_css_class("bold")
        heading.set_xalign(0)
        self.prepend(heading)

        omemo_trust_manager = OMEMOTrustManager(account)
        omemo_trust_manager.set_margin_top(18)
        self.append(omemo_trust_manager)


class ConnectionPage(GenericSettingPage):

    name = "connection"

    def __init__(self, account: str) -> None:

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Proxy"),
                SettingType.ACCOUNT_CONFIG,
                "proxy",
                name="proxy",
                props={
                    "data": self._get_proxies(),
                    "button-icon-name": "preferences-system-symbolic",
                    "button-callback": self._on_proxy_edit,
                },
            ),
            Setting(
                SettingKind.HOSTNAME,
                _("Hostname"),
                SettingType.DIALOG,
                desc=_("Manually set the hostname for the server"),
                props={"dialog": CutstomHostnameDialog},
            ),
            Setting(
                SettingKind.ENTRY, _("Resource"), SettingType.ACCOUNT_CONFIG, "resource"
            ),
            Setting(
                SettingKind.PRIORITY,
                _("Priority"),
                SettingType.DIALOG,
                props={"dialog": PriorityDialog},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use Unencrypted Connection"),
                SettingType.ACCOUNT_CONFIG,
                "use_plain_connection",
                desc=_("Use an unencrypted connection to the server"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Confirm Unencrypted Connection"),
                SettingType.ACCOUNT_CONFIG,
                "confirm_unencrypted_connection",
                desc=_("Show a confirmation dialog before connecting unencrypted"),
            ),
        ]
        GenericSettingPage.__init__(self, account, settings)

    @staticmethod
    def _get_proxies() -> dict[str, str]:
        proxies = {"": _("System")}
        proxies.update({proxy: proxy for proxy in app.settings.get_proxies()})
        proxies["no-proxy"] = _("No Proxy")
        return proxies

    @staticmethod
    def _on_proxy_edit(*args: Any) -> None:
        open_window("ManageProxies")

    def update_proxy_entries(self) -> None:
        dropdown_row = cast(DropDownSetting, self.listbox.get_setting("proxy"))
        dropdown_row.update_entries(self._get_proxies())


class AdvancedPage(GenericSettingPage):

    name = "advanced"

    def __init__(self, account: str) -> None:

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Contact Information"),
                SettingType.ACCOUNT_CONFIG,
                "request_user_data",
                desc=_("Request contact information (Tune, Location)"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Accept all Contact Requests"),
                SettingType.ACCOUNT_CONFIG,
                "autoauth",
                desc=_("Automatically accept all contact requests"),
            ),
            # TODO Jingle FT
            # Setting(SettingKind.DROPDOWN,
            #         _('Filetransfer Preference'),
            #         SettingType.ACCOUNT_CONFIG,
            #         'filetransfer_preference',
            #         props={'data': {'httpupload': _('Upload Files'),
            #                         'jingle': _('Send Files Directly')}},
            #         desc=_('Preferred file transfer mechanism for '
            #                'file drag&drop on a chat window')),
            Setting(
                SettingKind.SWITCH,
                _("Security Labels"),
                SettingType.ACCOUNT_CONFIG,
                "enable_security_labels",
                desc=_(
                    "Show labels describing confidentiality of "
                    "messages, if the server supports XEP-0258"
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Synchronize joined group chats"),
                SettingType.ACCOUNT_CONFIG,
                "autojoin_sync",
                desc=_("Synchronize joined group chats with other devices."),
            ),
        ]
        GenericSettingPage.__init__(self, account, settings)


class PriorityDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        neg_priority = app.settings.get("enable_negative_priority")
        if neg_priority:
            range_ = (-128, 127, 1)
        else:
            range_ = (0, 127, 1)

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Adjust to status"),
                SettingType.ACCOUNT_CONFIG,
                "adjust_priority_with_status",
            ),
            Setting(
                SettingKind.SPIN,
                _("Priority"),
                SettingType.ACCOUNT_CONFIG,
                "priority",
                bind="account::adjust_priority_with_status",
                inverted=True,
                props={"range_": range_},
            ),
        ]

        SettingsDialog.__init__(
            self, parent, _("Priority"), Gtk.DialogFlags.MODAL, settings, account
        )

    def _cleanup(self) -> None:
        # Update priority
        if self.account not in app.settings.get_active_accounts():
            return
        client = app.get_client(self.account)
        show = client.status
        status = client.status_message
        client.change_status(show, status)


class CutstomHostnameDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        type_values = ["START TLS", "DIRECT TLS", "PLAIN"]

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Enable"),
                SettingType.ACCOUNT_CONFIG,
                "use_custom_host",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Hostname"),
                SettingType.ACCOUNT_CONFIG,
                "custom_host",
                bind="account::use_custom_host",
            ),
            Setting(
                SettingKind.SPIN,
                _("Port"),
                SettingType.ACCOUNT_CONFIG,
                "custom_port",
                bind="account::use_custom_host",
                props={"range_": (0, 65535, 1)},
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Type"),
                SettingType.ACCOUNT_CONFIG,
                "custom_type",
                bind="account::use_custom_host",
                props={"data": type_values},
            ),
        ]

        SettingsDialog.__init__(
            self,
            parent,
            _("Connection Settings"),
            Gtk.DialogFlags.MODAL,
            settings,
            account,
        )


class CertificateDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(
                SettingKind.FILECHOOSER,
                _("Client Certificate"),
                SettingType.ACCOUNT_CONFIG,
                "client_cert",
                props={
                    "filefilter": [
                        Filter(name=_("All files"), patterns=["*"]),
                        Filter(name=_("PKCS12 Files"), patterns=["*.p12"]),
                    ]
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Encrypted Certificate"),
                SettingType.ACCOUNT_CONFIG,
                "client_cert_encrypted",
            ),
        ]

        SettingsDialog.__init__(
            self,
            parent,
            _("Certificate Settings"),
            Gtk.DialogFlags.MODAL,
            settings,
            account,
        )


class LoginDialog(SettingsDialog):
    def __init__(self, account: str, parent: Gtk.Window) -> None:

        settings = [
            Setting(
                SettingKind.ENTRY,
                _("Password"),
                SettingType.ACCOUNT_CONFIG,
                "password",
                bind="account::savepass",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Save Password"),
                SettingType.ACCOUNT_CONFIG,
                "savepass",
                enabled_func=(
                    lambda: not app.settings.get("use_keyring")
                    or passwords.is_keyring_available()
                ),
            ),
            Setting(
                SettingKind.CHANGEPASSWORD,
                _("Change Password"),
                SettingType.DIALOG,
                callback=self.on_password_change,
                props={"dialog": None},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use GSSAPI"),
                SettingType.ACCOUNT_CONFIG,
                "enable_gssapi",
            ),
        ]

        SettingsDialog.__init__(
            self, parent, _("Login Settings"), Gtk.DialogFlags.MODAL, settings, account
        )

    def on_password_change(self, new_password: str, _data: Any) -> None:
        try:
            new_password = saslprep(new_password)
        except Exception:
            # TODO: Show warning for invalid passwords
            return

        passwords.save_password(self.account, new_password)

    def _cleanup(self) -> None:
        savepass = app.settings.get_account_setting(self.account, "savepass")
        if not savepass:
            passwords.delete_password(self.account)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
import os
import re
import uuid
from collections.abc import Callable
from concurrent.futures import Future
from dataclasses import dataclass
from dataclasses import field
from functools import partial
from pathlib import Path
from urllib.parse import ParseResult
from urllib.parse import urlparse

from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.const import HTTPRequestError
from nbxmpp.http import HTTPRequest
from nbxmpp.util import convert_tls_error_flags

from gajim.common import app
from gajim.common import configpaths
from gajim.common import regex
from gajim.common.const import MIME_TYPES
from gajim.common.helpers import load_file_async
from gajim.common.helpers import write_file_async
from gajim.common.i18n import _
from gajim.common.multiprocess.thumbnail import create_thumbnail
from gajim.common.storage.archive import models as mod
from gajim.common.util.http import create_http_request
from gajim.common.util.preview import aes_decrypt
from gajim.common.util.preview import filename_from_uri
from gajim.common.util.preview import get_image_paths
from gajim.common.util.preview import get_previewable_mime_types
from gajim.common.util.preview import guess_mime_type
from gajim.common.util.preview import parse_fragment
from gajim.common.util.preview import split_geo_uri
from gajim.common.util.user_strings import get_tls_error_phrases

log = logging.getLogger('gajim.c.preview')

IRI_RX = re.compile(regex.IRI)

PREVIEWABLE_MIME_TYPES = get_previewable_mime_types()
mime_types = set(MIME_TYPES)
# Merge both: if it’s a previewable image, it should be allowed
ALLOWED_MIME_TYPES = mime_types.union(PREVIEWABLE_MIME_TYPES)

AudioSampleT = list[tuple[float, float]]


@dataclass
class AudioPreviewState:
    duration: float = 0.0
    position: float = 0.0
    is_eos: bool = False
    speed: float = 1.0
    is_timestamp_positive: bool = True
    samples: AudioSampleT = field(default_factory=list)
    is_audio_analyzed = False


class Preview:
    def __init__(self,
                 uri: str,
                 urlparts: ParseResult | None,
                 orig_path: Path | None,
                 thumb_path: Path | None,
                 size: int,
                 widget: Any,
                 from_us: bool = False,
                 context: str | None = None
                 ) -> None:

        self.id = str(uuid.uuid4())
        self._uri = uri
        self._urlparts = urlparts
        self._filename = filename_from_uri(uri)
        self._widget = widget
        self._from_us = from_us
        self._context = context

        self.account = widget.account
        self.orig_path = orig_path
        self.thumb_path = thumb_path
        self.size = size * app.window.get_scale_factor()

        self.thumbnail: bytes | None = None
        self.mime_type: str = ''
        self.file_size: int = 0
        self.download_in_progress = False

        self.info_message: str | None = None

        self._request = None

        self.key: bytes | None = None
        self.iv: bytes | None = None
        if self.is_aes_encrypted and urlparts is not None:
            try:
                self.key, self.iv = parse_fragment(urlparts.fragment)
            except ValueError as err:
                log.error('Parsing fragment for AES decryption '
                          'failed: %s', err)

    @property
    def is_geo_uri(self) -> bool:
        return self._uri.startswith('geo:')

    @property
    def is_web_uri(self) -> bool:
        return not self.is_geo_uri

    @property
    def is_previewable(self) -> bool:
        return self.mime_type in PREVIEWABLE_MIME_TYPES

    @property
    def is_audio(self) -> bool:
        is_allowed = bool(self.mime_type in ALLOWED_MIME_TYPES)
        return is_allowed and self.mime_type.startswith('audio/')

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def from_us(self) -> bool:
        return self._from_us

    @property
    def context(self) -> str | None:
        return self._context

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def request(self) -> HTTPRequest | None:
        return self._request

    @property
    def request_uri(self) -> str:
        if self._urlparts is None:
            return ''
        if self.is_aes_encrypted:
            # Remove fragments so we dont transmit it to the server
            urlparts = self._urlparts._replace(scheme='https', fragment='')
            return urlparts.geturl()
        return self._urlparts.geturl()

    @property
    def is_aes_encrypted(self) -> bool:
        if self._urlparts is None:
            return False
        return self._urlparts.scheme == 'aesgcm'

    @property
    def thumb_exists(self) -> bool:
        if self.thumb_path is None:
            return False
        return self.thumb_path.exists()

    @property
    def orig_exists(self) -> bool:
        if self.orig_path is None:
            return False
        return self.orig_path.exists()

    def update_widget(self, data: bytes | None = None) -> None:
        self._widget.update(self, data)

    def update_progress(self, progress: float, request: HTTPRequest) -> None:
        self._request = request
        self._widget.update_progress(self, progress)


class PreviewManager:
    def __init__(self) -> None:
        self._orig_dir = configpaths.get('DOWNLOADS')
        self._thumb_dir = configpaths.get('DOWNLOADS_THUMB')

        self._previews: dict[str, Preview] = {}

        # Holds active audio preview sessions
        # for resuming after switching chats
        self._audio_sessions: dict[int, AudioPreviewState] = {}

        # References a stop function for each audio preview, which allows us
        # to stop previews by preview_id, see stop_audio_except(preview_id)
        self._audio_stop_functions: dict[int, Callable[..., None]] = {}

        log.info('Supported mime types for preview')
        log.info(sorted(PREVIEWABLE_MIME_TYPES))

    def get_preview(self, preview_id: str) -> Preview | None:
        return self._previews.get(preview_id)

    def clear_previews(self) -> None:
        self._previews.clear()

    def get_audio_state(self,
                        preview_id: int
                        ) -> AudioPreviewState:

        state = self._audio_sessions.get(preview_id)
        if state is not None:
            return state
        self._audio_sessions[preview_id] = AudioPreviewState()
        return self._audio_sessions[preview_id]

    def register_audio_stop_func(self,
                                 preview_id: int,
                                 stop_func: Callable[..., None]
                                 ) -> None:

        self._audio_stop_functions[preview_id] = stop_func

    def unregister_audio_stop_func(self, preview_id: int) -> None:
        self._audio_stop_functions.pop(preview_id, None)

    def stop_audio_except(self, preview_id: int) -> None:
        # Stops playback of all audio previews except of for preview_id.
        # This makes sure that only one preview is played at the time.
        for id_, stop_func in self._audio_stop_functions.items():
            if id_ != preview_id:
                stop_func()

    @staticmethod
    def _accept_uri(urlparts: ParseResult,
                    uri: str,
                    oob_url: str | None
                    ) -> bool:

        # geo
        if urlparts.scheme == 'geo':
            return True

        if not urlparts.netloc:
            return False

        # aesgcm
        if urlparts.scheme == 'aesgcm':
            return True

        # http/https
        if urlparts.scheme in ('https', 'http'):
            if app.settings.get('preview_allow_all_images'):
                mime_type = guess_mime_type(uri)
                if mime_type not in MIME_TYPES:
                    log.info('%s not in allowed mime types', mime_type)
                    return False

                if mime_type == 'application/octet-stream' and uri != oob_url:  # noqa: SIM103
                    # guess_mime_type yields 'application/octet-stream' for
                    # paths without suffix. Check oob_url to make sure we
                    # display a preview for files sent via http upload.
                    return False
                return True

            if oob_url is None:
                log.info('No oob url for: %s', uri)
                return False

            if uri != oob_url:
                log.info('uri != oob url: %s != %s', uri, oob_url)
                return False
            return True

        log.info('Unsupported URI scheme: %s', uri)
        return False

    def is_previewable(self,
                       text: str,
                       oob_data: list[mod.OOB]
                       ) -> bool:


        if not IRI_RX.fullmatch(text):
            # urlparse removes whitespace (and who knows what else) from URLs,
            # so can't be used for validation.
            return False

        uri = text
        try:
            urlparts = urlparse(uri)
        except Exception:
            return False

        oob_url = None if not oob_data else oob_data[0].url
        if not self._accept_uri(urlparts, uri, oob_url):
            return False

        if urlparts.scheme == 'geo':
            try:
                split_geo_uri(uri)
            except Exception as err:
                log.info('Bad geo URI %s: %s', uri, err)
                return False

        return True

    def create_preview(self,
                       uri: str,
                       widget: Any,
                       from_us: bool,
                       context: str | None = None
                       ) -> None:

        if uri.startswith('geo:'):
            preview = Preview(uri, None, None, None, 96, widget)
            preview.update_widget()
            self._previews[preview.id] = preview
            return

        preview = self._process_web_uri(uri, widget, from_us, context)
        self._previews[preview.id] = preview

        if not app.settings.get('enable_file_preview'):
            preview.update_widget()
            return

        if not preview.orig_exists:
            if context is not None and not from_us:
                allow_in_public = app.settings.get('preview_anonymous_muc')
                if context == 'public' and not allow_in_public:
                    preview.update_widget()
                    return

            self.download_content(preview)

        elif not preview.thumb_exists:
            assert preview.orig_path is not None
            load_file_async(preview.orig_path,
                            self._on_orig_load_finished,
                            preview)

        else:
            assert preview.thumb_path is not None
            load_file_async(preview.thumb_path,
                            self._on_thumb_load_finished,
                            preview)

    def create_thumbnail(self, preview: Preview, data: bytes) -> None:
        future = app.process_pool.submit(
            create_thumbnail,
            data,
            preview.size,
            preview.mime_type,
        )
        future.add_done_callback(partial(self._write_thumbnail, preview))

    def _write_thumbnail(self, preview: Preview, future: Future[bytes | None]) -> None:
        try:
            result = future.result()
            if result is None:
                raise ValueError("result is None")
        except Exception as error:
            preview.info_message = _('Creating thumbnail failed')
            preview.update_widget()
            log.warning('Creating thumbnail failed for: %s %s',
                        preview.orig_path, error)
            return

        preview.thumbnail = result

        assert preview.thumb_path is not None
        write_file_async(preview.thumb_path,
                         preview.thumbnail,
                         self._on_thumb_write_finished,
                         preview)

    def _process_web_uri(self,
                         uri: str,
                         widget: Any,
                         from_us: bool,
                         context: str | None = None
                         ) -> Preview:
        urlparts = urlparse(uri)
        size = app.settings.get('preview_size')
        orig_path, thumb_path = get_image_paths(uri,
                                                urlparts,
                                                size,
                                                self._orig_dir,
                                                self._thumb_dir)
        return Preview(uri,
                       urlparts,
                       orig_path,
                       thumb_path,
                       size,
                       widget,
                       from_us,
                       context=context)

    def _on_orig_load_finished(self,
                               data: bytes | None,
                               error: GLib.Error | None,
                               preview: Preview) -> None:
        if preview.thumb_path is None or preview.orig_path is None:
            return

        if data is None:
            log.error('%s: %s', preview.orig_path.name, error)
            return

        preview.mime_type = guess_mime_type(preview.orig_path, data)
        preview.file_size = os.path.getsize(preview.orig_path)

        if preview.is_previewable:
            self.create_thumbnail(preview, data)
        else:
            preview.update_widget()

    @staticmethod
    def _on_thumb_load_finished(data: bytes | None,
                                error: GLib.Error | None,
                                preview: Preview) -> None:

        if preview.thumb_path is None or preview.orig_path is None:
            return

        if data is None:
            log.error('%s: %s', preview.thumb_path.name, error)
            return

        preview.thumbnail = data
        # Thumbnails are stored always as PNG, we don’t know the
        # mime-type of the original picture
        preview.mime_type = guess_mime_type(preview.orig_path)
        preview.file_size = os.path.getsize(preview.orig_path)

        preview.update_widget(data=data)

    def download_content(self,
                         preview: Preview,
                         force: bool = False
                         ) -> None:

        if preview.download_in_progress:
            log.info('Download already in progress')
            return

        if preview.account is None:
            # History Window can be opened without account context
            # This means we can not apply proxy settings
            return

        log.info('Start downloading: %s', preview.request_uri)
        preview.download_in_progress = True

        request = create_http_request(preview.account)
        request.set_user_data(preview)
        request.connect('accept-certificate', self._accept_certificate)
        request.connect('content-sniffed', self._on_content_sniffed, force)
        request.connect('response-progress', self._on_response_progress)

        request.send('GET', preview.request_uri, callback=self._on_finished)

    def _accept_certificate(self,
                            request: HTTPRequest,
                            _certificate: Gio.TlsCertificate,
                            certificate_errors: Gio.TlsCertificateFlags,
                            ) -> bool:

        if not app.settings.get('preview_verify_https'):
            return True

        phrases = get_tls_error_phrases(
            convert_tls_error_flags(certificate_errors))
        log.warning(
            'TLS verification failed: %s (0x%02x)', phrases, certificate_errors)

        preview = cast(Preview, request.get_user_data())
        preview.info_message = _('TLS verification failed: %s') % phrases[0]
        preview.download_in_progress = False
        preview.update_widget()
        return False

    def _on_content_sniffed(self,
                            request: HTTPRequest,
                            content_length: int,
                            content_type: str,
                            force: bool
                            ) -> None:

        uri = request.get_uri()
        assert uri is not None
        uri = uri.to_string()
        preview = cast(Preview, request.get_user_data())
        preview.mime_type = content_type
        preview.file_size = content_length

        if content_type not in ALLOWED_MIME_TYPES and not force:
            log.info('Not an allowed content type: %s, %s', content_type, uri)
            request.cancel()
            preview.download_in_progress = False
            return

        if content_length > int(app.settings.get('preview_max_file_size')):
            log.info(
                'File size (%s) too big for URL: "%s"',
                content_length, uri)
            if force:
                preview.info_message = None
            else:
                request.cancel()
                preview.download_in_progress = False
                preview.info_message = _('Automatic preview disabled '
                                         '(file too big)')

        preview.update_widget()

    def _on_response_progress(self,
                              request: HTTPRequest,
                              progress: float,
                              ) -> None:

        preview = cast(Preview, request.get_user_data())
        preview.update_progress(progress, request)

    def _on_finished(self, request: HTTPRequest) -> None:

        preview = cast(Preview, request.get_user_data())
        preview.download_in_progress = False

        if not request.is_complete():
            error = request.get_error_string()
            log.warning('Download failed: %s - %s', preview.request_uri, error)
            if request.get_error() != HTTPRequestError.CANCELLED:
                preview.info_message = _('Download failed (%s)') % error
            preview.update_widget()
            return

        preview.info_message = None

        data = request.get_data()
        if not data:
            return

        if preview.is_aes_encrypted:
            if preview.key is not None and preview.iv is not None:
                try:
                    data = aes_decrypt(preview.key, preview.iv, data)
                except Exception as error:
                    log.exception('Decryption failed')
                    preview.info_message = _('Decryption failed')
                    preview.update_widget()
                    return

        if preview.mime_type == 'application/octet-stream':
            if preview.orig_path is not None:
                preview.mime_type = guess_mime_type(preview.orig_path, data)

        assert preview.orig_path is not None
        write_file_async(preview.orig_path,
                         data,
                         self._on_orig_write_finished,
                         preview)

        if not app.settings.get('enable_file_preview'):
            preview.update_widget()
            return

        if preview.is_previewable:
            self.create_thumbnail(preview, data)

    @staticmethod
    def _on_orig_write_finished(_result: bool,
                                error: GLib.Error | None,
                                preview: Preview) -> None:
        if preview.orig_path is None:
            return

        if error is not None:
            log.error('%s: %s', preview.orig_path.name, error)
            return

        log.info('File stored: %s', preview.orig_path.name)
        preview.file_size = os.path.getsize(preview.orig_path)
        if not preview.is_previewable:
            # Don’t update preview if thumb is already displayed,
            # but update preview for audio files
            preview.update_widget()

    @staticmethod
    def _on_thumb_write_finished(_result: bool,
                                 error: GLib.Error | None,
                                 preview: Preview) -> None:
        if preview.thumb_path is None:
            return

        if error is not None:
            log.error('%s: %s', preview.thumb_path.name, error)
            if not preview.thumb_exists:
                # Generating a preview can fail if the file already exists
                # Only abort if thumbnail has not been stored in preview
                return

        log.info('Thumbnail stored: %s ', preview.thumb_path.name)

        if preview.thumbnail is None:
            return

        preview.update_widget(data=preview.thumbnail)

    def cancel_download(self, preview: Preview) -> None:
        assert preview.request is not None
        preview.request.cancel()
        preview.download_in_progress = False



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import json
import sys
from dataclasses import dataclass
from pathlib import Path

from packaging.requirements import Requirement
from packaging.version import Version as V

import gajim

from .helpers import is_shipped_plugin
from .plugins_i18n import _ as p_

BLOCKED_PLUGINS = [
    'appindicator_integration',
    'omemo',
    'plugin_installer',
    'syntax_highlight',
    'url_image_preview'
]

GAJIM_VERSION = V(gajim.__version__.split('+', 1)[0]).base_version


@dataclass(frozen=True, eq=True)
class PluginManifest:
    name: str
    short_name: str
    description: str
    authors: list[str]
    homepage: str
    config_dialog: bool
    version: V
    requirements: list[Requirement]
    platforms: list[str]
    path: Path | None = None

    def __hash__(self):
        return hash(f'{self.short_name}_{self.version}')

    @property
    def is_usable(self) -> bool:
        if self.short_name in BLOCKED_PLUGINS:
            return False

        if not self._check_requirements():
            return False

        platform = sys.platform
        if platform not in ('win32', 'darwin', 'linux'):
            # sys.platform can return an unknown amount of unix/linux derivates
            platform = 'others'

        return platform in self.platforms

    @property
    def is_shipped(self) -> bool:
        if self.path is None:
            return False
        return is_shipped_plugin(self.path)

    def get_remote_url(self, repository_url: str) -> str:
        filename = f'{self.short_name}_{self.version}.zip'
        return f'{repository_url}/{self.short_name}/{filename}'

    @classmethod
    def from_path(cls, path: Path) -> PluginManifest:
        manifest_path = path / 'plugin-manifest.json'
        if not manifest_path.exists():
            raise ValueError(f'Not a plugin path: {path}')

        if manifest_path.is_dir():
            raise ValueError(f'Not a plugin path: {path}')

        with manifest_path.open(encoding='utf8') as f:
            try:
                manifest = json.load(f)
            except Exception as error:
                raise ValueError(f'Error while parsing manifest: '
                                 f'{path}, {error}')
        return cls.from_manifest_json(manifest, manifest_path)

    def _check_requirements(self) -> bool:
        return any(GAJIM_VERSION in req.specifier for req in self.requirements)

    @classmethod
    def from_manifest_json(cls,
                           manifest: dict[str, Any],
                           path: Path | None = None) -> PluginManifest:
        return cls(
            name=manifest['name'],
            short_name=manifest['short_name'],
            description=p_(manifest['description']),
            authors=manifest['authors'],
            homepage=manifest['homepage'],
            config_dialog=manifest['config_dialog'],
            version=V(manifest['version']),
            requirements=[Requirement(r) for r in manifest['requirements']],
            platforms=manifest['platforms'],
            path=path.parent if path is not None else path)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

from collections.abc import Callable

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp import JID

from gajim.common import app
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.events import MucDecline
from gajim.common.events import MucInvitation
from gajim.common.events import Notification
from gajim.common.events import SubscribePresenceReceived
from gajim.common.events import UnsubscribedPresenceReceived
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.util.muc import get_groupchat_name

from gajim.gtk.menus import get_subscription_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_listbox_row_count
from gajim.gtk.util.misc import iterate_listbox_children
from gajim.gtk.util.window import open_window

NotificationActionListT = list[
    tuple[str, Callable[[Gio.SimpleAction, GLib.Variant], Any], str | None]
]


class NotificationManager(Gtk.ListBox, SignalManager):
    def __init__(self, account: str) -> None:
        Gtk.ListBox.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self.set_selection_mode(Gtk.SelectionMode.NONE)
        self.set_halign(Gtk.Align.CENTER)
        self.add_css_class("notification-listbox")

        label = Gtk.Label(label=_("No Notifications"))
        label.set_valign(Gtk.Align.START)
        label.add_css_class("dim-label")
        self.set_placeholder(label)

        self._add_actions()

    def do_unroot(self) -> None:
        self._disconnect_all()
        self._remove_actions()
        self._client.disconnect_all_from_obj(self)
        Gtk.ListBox.do_unroot(self)
        app.check_finalize(self)

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        self.update_actions()

    def _add_actions(self) -> None:
        actions: NotificationActionListT = [
            ("subscription-accept", self._on_subscription_accept, "as"),
            ("subscription-deny", self._on_subscription_deny, "s"),
            ("subscription-deny-all", self._on_subscription_deny_all, None),
            ("subscription-block", self._on_subscription_block, "s"),
            ("subscription-report", self._on_subscription_report, "s"),
        ]
        for action in actions:
            action_name, func, typ = action
            if typ is not None:
                typ = GLib.VariantType.new(typ)
            act = Gio.SimpleAction.new(f"{action_name}-{self._account}", typ)
            self._connect(act, "activate", func)
            app.window.add_action(act)

    def update_actions(self) -> None:
        online = app.account_is_connected(self._account)
        blocking_support = self._client.get_module("Blocking").supported

        sub_accept = app.window.get_action(f"subscription-accept-{self._account}")
        sub_accept.set_enabled(online)

        sub_deny = app.window.get_action(f"subscription-deny-{self._account}")
        sub_deny.set_enabled(online)

        sub_deny_all = app.window.get_action(f"subscription-deny-all-{self._account}")
        sub_deny_all.set_enabled(online)

        sub_block = app.window.get_action(f"subscription-block-{self._account}")
        sub_block.set_enabled(online and blocking_support)

        sub_report = app.window.get_action(f"subscription-report-{self._account}")
        sub_report.set_enabled(online and blocking_support)

    def _remove_actions(self) -> None:
        actions = [
            "subscription-accept",
            "subscription-deny",
            "subscription-deny-all",
            "subscription-block",
            "subscription-report",
        ]
        for action in actions:
            app.window.remove_action(f"{action}-{self._account}")

    def update_unread_count(self):
        count = get_listbox_row_count(self)
        app.window.update_account_unread_count(self._account, count)

    def remove_row(self, row: NotificationRow) -> None:
        self.remove(row)
        self.update_unread_count()

    def _on_subscription_accept(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        jid, nickname = param.get_strv()
        row = self._get_notification_row(jid)
        self._client.get_module("Presence").subscribed(jid)
        jid = JID.from_string(jid)
        contact = self._client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, BareContact)
        if not contact.is_in_roster:
            open_window(
                "AddContact",
                account=self._account,
                jid=jid,
                nick=nickname or contact.name,
            )
        if row is not None:
            self.remove(row)
            self.update_unread_count()

    def _on_subscription_block(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        jid = param.get_string()
        self._deny_request(jid)
        self._client.get_module("Blocking").block([jid])
        row = self._get_notification_row(jid)
        if row is not None:
            self.remove(row)
            self.update_unread_count()

    def _on_subscription_report(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        jid = param.get_string()
        self._deny_request(jid)
        self._client.get_module("Blocking").block([jid], report="spam")
        row = self._get_notification_row(jid)
        if row is not None:
            self.remove(row)
            self.update_unread_count()

    def _on_subscription_deny(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        jid = param.get_string()
        self._deny_request(jid)
        row = self._get_notification_row(jid)
        if row is not None:
            self.remove(row)
            self.update_unread_count()

    def _on_subscription_deny_all(
        self, _action: Gio.SimpleAction, _param: GLib.Variant
    ) -> None:

        for row in cast(list[NotificationRow], iterate_listbox_children(self)):
            if row.type != "subscribe":
                continue
            self._deny_request(row.jid)
            self.remove(row)
        self.update_unread_count()

    def _deny_request(self, jid: str) -> None:
        self._client.get_module("Presence").unsubscribed(jid)

    def _get_notification_row(self, jid: str) -> NotificationRow | None:
        rows = cast(list[NotificationRow], iterate_listbox_children(self))
        for row in rows:
            if row.jid == jid:
                return row
        return None

    def add_subscription_request(self, event: SubscribePresenceReceived) -> None:
        row = self._get_notification_row(event.jid)
        if row is None:
            new_row = SubscriptionRequestRow(
                self._account, event.jid, event.status, event.user_nick
            )
            self.append(new_row)

            nick = event.user_nick
            if not nick:
                contact = self._client.get_module("Contacts").get_contact(event.jid)
                assert isinstance(contact, BareContact)
                nick = contact.name
            text = _("%s asks you to share your status") % nick

            app.ged.raise_event(
                Notification(
                    account=self._account,
                    jid=event.jid,
                    type="subscription-request",
                    title=_("Subscription Request"),
                    text=text,
                )
            )
        elif row.type == "unsubscribed":
            self.remove(row)

        self.update_unread_count()

    def add_unsubscribed(self, event: UnsubscribedPresenceReceived) -> None:
        row = self._get_notification_row(event.jid)
        if row is None:
            new_row = UnsubscribedRow(self._account, event.jid)
            self.append(new_row)
            self.update_unread_count()

            contact = self._client.get_module("Contacts").get_contact(event.jid)
            assert isinstance(contact, BareContact)
            text = _("%s stopped sharing their status") % contact.name

            app.ged.raise_event(
                Notification(
                    account=self._account,
                    jid=event.jid,
                    type="unsubscribed",
                    title=_("Contact Unsubscribed"),
                    text=text,
                )
            )
        elif row.type == "subscribe":
            self.remove(row)

        self.update_unread_count()

    def add_invitation_received(self, event: MucInvitation) -> None:
        row = self._get_notification_row(str(event.muc))
        if row is not None:
            return

        jid = event.from_.bare
        client = app.get_client(event.account)
        muc_contact = client.get_module("Contacts").get_contact(event.muc)
        assert isinstance(muc_contact, GroupchatContact)

        new_row = InvitationReceivedRow(self._account, event, muc_contact)
        self.append(new_row)
        self.update_unread_count()

        if muc_contact.muc_context == "private" and not event.muc.bare_match(
            event.from_
        ):
            contact = self._client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            text = _("%(contact)s invited you to %(chat)s") % {
                "contact": contact.name,
                "chat": event.info.muc_name,
            }
        else:
            text = _("You have been invited to %s") % event.info.muc_name

        app.ged.raise_event(
            Notification(
                account=self._account,
                jid=jid,
                type="group-chat-invitation",
                title=_("Group Chat Invitation"),
                text=text,
            )
        )

    def add_invitation_declined(self, event: MucDecline) -> None:
        row = self._get_notification_row(str(event.muc))
        if row is not None:
            return

        new_row = InvitationDeclinedRow(self._account, event)
        self.append(new_row)
        self.update_unread_count()


class NotificationRow(Gtk.ListBoxRow, SignalManager):
    def __init__(self, account: str, jid: str) -> None:
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self.jid = jid
        self.type = ""

        self.grid = Gtk.Grid(column_spacing=12)
        self.set_child(self.grid)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    @staticmethod
    def _generate_label() -> Gtk.Label:
        label = Gtk.Label()
        label.set_halign(Gtk.Align.START)
        label.set_hexpand(True)
        label.set_xalign(0)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_max_width_chars(30)
        return label

    def _generate_avatar_image(self, jid: str | JID) -> Gtk.Image:
        contact = self._client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        if isinstance(contact, GroupchatContact):
            texture = contact.get_avatar(AvatarSize.ROSTER, self.get_scale_factor())
        else:
            texture = contact.get_avatar(
                AvatarSize.ROSTER, self.get_scale_factor(), add_show=False
            )
        image = Gtk.Image.new_from_paintable(texture)
        image.set_valign(Gtk.Align.CENTER)
        image.set_pixel_size(AvatarSize.ROSTER)
        return image


class SubscriptionRequestRow(NotificationRow):
    def __init__(
        self, account: str, jid: str, text: str, user_nick: str | None = None
    ) -> None:
        NotificationRow.__init__(self, account, jid)
        self.type = "subscribe"

        image = self._generate_avatar_image(jid)
        self.grid.attach(image, 1, 1, 1, 2)

        if user_nick is not None:
            escaped_nick = GLib.markup_escape_text(user_nick)
            nick_markup = f"<b>{escaped_nick}</b> ({jid})"
        else:
            nick_markup = f"<b>{jid}</b>"

        nick_label = self._generate_label()
        nick_label.set_tooltip_markup(nick_markup)
        nick_label.set_markup(nick_markup)
        self.grid.attach(nick_label, 2, 1, 1, 1)

        message_text = GLib.markup_escape_text(text)
        text_label = self._generate_label()
        text_label.set_text(message_text)
        text_label.set_tooltip_text(message_text)
        text_label.add_css_class("dim-label")
        self.grid.attach(text_label, 2, 2, 1, 1)

        accept_button = Gtk.Button.new_with_label(label=_("Accept"))
        accept_button.set_valign(Gtk.Align.CENTER)
        accept_button.set_action_name(f"win.subscription-accept-{self._account}")
        accept_button.set_action_target_value(
            GLib.Variant.new_strv([self.jid, user_nick or ""])
        )
        self.grid.attach(accept_button, 3, 1, 1, 2)

        more_image = Gtk.Image.new_from_icon_name("view-more-symbolic")
        more_button = Gtk.MenuButton()
        more_button.set_valign(Gtk.Align.CENTER)
        more_button.set_child(more_image)
        subscription_menu = get_subscription_menu(
            self._account, JID.from_string(self.jid)
        )
        more_button.set_menu_model(subscription_menu)
        self.grid.attach(more_button, 4, 1, 1, 2)


class UnsubscribedRow(NotificationRow):
    def __init__(self, account: str, jid: str) -> None:
        NotificationRow.__init__(self, account, jid)
        self.type = "unsubscribed"

        image = self._generate_avatar_image(jid)
        self.grid.attach(image, 1, 1, 1, 2)

        contact = self._client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, BareContact)
        nick_markup = f"<b>{contact.name}</b>"
        nick_label = self._generate_label()
        nick_label.set_tooltip_markup(nick_markup)
        nick_label.set_markup(nick_markup)
        self.grid.attach(nick_label, 2, 1, 1, 1)

        message_text = _("Stopped sharing their status with you")
        text_label = self._generate_label()
        text_label.set_text(message_text)
        text_label.set_tooltip_text(message_text)
        text_label.add_css_class("dim-label")
        self.grid.attach(text_label, 2, 2, 1, 1)

        remove_button = Gtk.Button.new_with_label(label=_("Remove"))
        remove_button.set_valign(Gtk.Align.CENTER)
        remove_button.set_tooltip_text(_("Remove from contact list"))
        remove_button.set_action_name(f"win.{self._account}-remove-contact")
        remove_button.set_action_target_value(
            GLib.Variant("as", [self._account, str(self.jid)])
        )
        self.grid.attach(remove_button, 3, 1, 1, 2)

        dismiss_button = Gtk.Button.new_from_icon_name("window-close-symbolic")
        dismiss_button.set_valign(Gtk.Align.CENTER)
        dismiss_button.set_tooltip_text(_("Remove Notification"))
        self._connect(dismiss_button, "clicked", self._on_dismiss)
        self.grid.attach(dismiss_button, 4, 1, 1, 2)

    def _on_dismiss(self, _button: Gtk.Button) -> None:
        listbox = cast(NotificationManager, self.get_parent())
        listbox.remove_row(self)


class InvitationReceivedRow(NotificationRow):
    def __init__(
        self, account: str, event: MucInvitation, muc_contact: GroupchatContact
    ) -> None:
        NotificationRow.__init__(self, account, str(event.muc))
        self.type = "invitation-received"

        self._muc_contact = muc_contact
        self._muc_contact.connect("room-joined", self._on_room_joined)

        self._event = event

        jid = event.from_.bare
        image = self._generate_avatar_image(jid)
        self.grid.attach(image, 1, 1, 1, 2)

        title_label = self._generate_label()
        title_label.set_text(_("Group Chat Invitation Received"))
        title_label.add_css_class("bold")
        self.grid.attach(title_label, 2, 1, 1, 1)

        if self._muc_contact.muc_context == "private" and not event.muc.bare_match(
            event.from_
        ):
            contact = self._client.get_module("Contacts").get_contact(jid)
            assert isinstance(contact, BareContact)
            invitation_text = _("%(contact)s invited you to %(chat)s") % {
                "contact": contact.name,
                "chat": event.info.muc_name,
            }
        else:
            invitation_text = _("You have been invited to %s") % event.info.muc_name
        text_label = self._generate_label()
        text_label.set_text(invitation_text)
        text_label.set_tooltip_text(invitation_text)
        text_label.add_css_class("dim-label")
        self.grid.attach(text_label, 2, 2, 1, 1)

        show_button = Gtk.Button.new_with_label(label=_("Show"))
        show_button.set_valign(Gtk.Align.CENTER)
        show_button.set_halign(Gtk.Align.END)
        show_button.set_tooltip_text(_("Show Invitation"))
        self._connect(show_button, "clicked", self._on_show_invitation)
        self.grid.attach(show_button, 3, 1, 1, 2)

        decline_button = Gtk.Button.new_with_label(label=_("Decline"))
        decline_button.set_valign(Gtk.Align.CENTER)
        decline_button.set_tooltip_text(_("Decline Invitation"))
        self._connect(decline_button, "clicked", self._on_decline_invitation)
        self.grid.attach(decline_button, 4, 1, 1, 2)

    def do_unroot(self) -> None:
        self._muc_contact.disconnect_all_from_obj(self)
        NotificationRow.do_unroot(self)

    def _on_show_invitation(self, _button: Gtk.Button) -> None:
        open_window(
            "GroupChatInvitationDialog", account=self._account, event=self._event
        )
        listbox = cast(NotificationManager, self.get_parent())
        listbox.remove_row(self)

    def _on_decline_invitation(self, _button: Gtk.Button) -> None:
        self._client.get_module("MUC").decline(self.jid, self._event.from_)
        listbox = cast(NotificationManager, self.get_parent())
        listbox.remove_row(self)

    def _on_room_joined(self, contact: GroupchatContact, signal_name: str) -> None:
        listbox = cast(NotificationManager, self.get_parent())
        listbox.remove_row(self)


class InvitationDeclinedRow(NotificationRow):
    def __init__(self, account: str, event: MucDecline) -> None:
        NotificationRow.__init__(self, account, str(event.muc))
        self.type = "invitation-declined"

        jid = event.from_.bare

        image = self._generate_avatar_image(jid)
        self.grid.attach(image, 1, 1, 1, 2)

        title_label = self._generate_label()
        title_label.set_text(_("Group Chat Invitation Declined"))
        title_label.add_css_class("bold")
        self.grid.attach(title_label, 2, 1, 1, 1)

        contact = self._client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, BareContact)
        muc_name = get_groupchat_name(self._client, event.muc)
        invitation_text = _("%(contact)s declined your invitation to %(chat)s") % {
            "contact": contact.name,
            "chat": muc_name,
        }
        text_label = self._generate_label()
        text_label.set_text(invitation_text)
        text_label.set_tooltip_text(invitation_text)
        text_label.add_css_class("dim-label")
        self.grid.attach(text_label, 2, 2, 1, 1)

        dismiss_button = Gtk.Button.new_from_icon_name("window-close-symbolic")
        dismiss_button.set_valign(Gtk.Align.CENTER)
        dismiss_button.set_halign(Gtk.Align.END)
        dismiss_button.set_tooltip_text(_("Remove Notification"))
        self._connect(dismiss_button, "clicked", self._on_dismiss)
        self.grid.attach(dismiss_button, 3, 1, 1, 2)

    def _on_dismiss(self, _button: Gtk.Button) -> None:
        listbox = cast(NotificationManager, self.get_parent())
        listbox.remove_row(self)



# Copyright (C) 2003-2017 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2004-2005 Vincent Hanquez <tab AT snarc.org>
# Copyright (C) 2005 Alex Podaras <bigpod AT gmail.com>
#                    Norman Rasmussen <norman AT rasmussen.co.za>
#                    Stéphan Kochen <stephan AT kochen.nl>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
#                         Alex Mauer <hawke AT hawkesnest.net>
# Copyright (C) 2005-2007 Travis Shirk <travis AT pobox.com>
#                         Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2006 Junglecow J <junglecow AT gmail.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2008 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Lukas Petrovicky <lukas AT petrovicky.net>
#                    James Newton <redshodan AT gmail.com>
# Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
#                         Julien Pivotto <roidelapluie AT gmail.com>
#                         Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2016-2017 Emmanuel Gil Peyrot <linkmauve AT linkmauve.fr>
#                         Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import os
import sys
from collections.abc import Callable
from datetime import datetime
from datetime import timedelta
from datetime import UTC
from urllib.parse import unquote

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.protocol import InvalidJid

import gajim
from gajim.common import app
from gajim.common import configpaths
from gajim.common import events
from gajim.common import ged
from gajim.common import idle
from gajim.common.application import CoreApplication
from gajim.common.const import GAJIM_FAQ_URI
from gajim.common.const import GAJIM_PRIVACY_POLICY_URI
from gajim.common.const import GAJIM_SUPPORT_JID
from gajim.common.const import GAJIM_WIKI_URI
from gajim.common.helpers import load_json
from gajim.common.i18n import _
from gajim.common.modules.contacts import ResourceContact
from gajim.common.util.uri import open_uri

from gajim.gtk import menus
from gajim.gtk import structs
from gajim.gtk.about import AboutDialog
from gajim.gtk.avatar import AvatarStorage
from gajim.gtk.const import ACCOUNT_ACTIONS
from gajim.gtk.const import ALWAYS_ACCOUNT_ACTIONS
from gajim.gtk.const import APP_ACTIONS
from gajim.gtk.const import FEATURE_ACCOUNT_ACTIONS
from gajim.gtk.const import MuteState
from gajim.gtk.const import ONLINE_ACCOUNT_ACTIONS
from gajim.gtk.dialogs import ConfirmationDialog
from gajim.gtk.dialogs import DialogButton
from gajim.gtk.dialogs import ShortcutsWindow
from gajim.gtk.menus import get_main_menu
from gajim.gtk.util.icons import get_icon_theme
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import get_app_windows
from gajim.gtk.util.window import open_window

ActionListT = list[tuple[str, Callable[[Gio.SimpleAction, GLib.Variant], Any]]]


class GajimApplication(Gtk.Application, CoreApplication):
    """Main class handling activation and command line."""

    def __init__(self):
        CoreApplication.__init__(self)
        flags = (
            Gio.ApplicationFlags.HANDLES_COMMAND_LINE
            | Gio.ApplicationFlags.CAN_OVERRIDE_APP_ID
        )
        Gtk.Application.__init__(
            self, application_id=app.get_default_app_id(), flags=flags
        )

        # required to track screensaver state
        self.props.register_session = True

        self.add_main_option(
            "version",
            ord("V"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show the application's version"),
        )

        self.add_main_option(
            "quiet",
            ord("q"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show only critical errors"),
        )

        self.add_main_option(
            "separate",
            ord("s"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _(
                "Separate profile files completely "
                "(even history database and plugins)"
            ),
        )

        self.add_main_option(
            "verbose",
            ord("v"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Print XML stanzas and other debug information"),
        )

        self.add_main_option(
            "profile",
            ord("p"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Use defined profile in configuration directory"),
            "NAME",
        )

        self.add_main_option(
            "config-path",
            ord("c"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Set configuration directory"),
            "PATH",
        )

        self.add_main_option(
            "loglevel",
            ord("l"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.STRING,
            _("Configure logging system"),
            "LEVEL",
        )

        self.add_main_option(
            "warnings",
            ord("w"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Show all warnings"),
        )

        self.add_main_option(
            "gdebug",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Sets an environment variable so GLib debug messages are printed"),
        )

        self.add_main_option(
            "cprofile",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Profile application with cprofile"),
        )

        self.add_main_option(
            "start-chat",
            0,
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Start a new chat"),
        )

        self.add_main_option_entries(self._get_remaining_entry())

        self.connect("activate", self._on_activate)
        self.connect("handle-local-options", self._handle_local_options)
        self.connect("command-line", self._command_line)
        self.connect("shutdown", self._shutdown)

    @staticmethod
    def _get_remaining_entry():
        option = GLib.OptionEntry()
        # https://gitlab.gnome.org/GNOME/pygobject/-/issues/608
        option.arg = int(GLib.OptionArg.STRING_ARRAY)  # pyright: ignore
        option.arg_data = None
        option.arg_description = "[URI …]"
        option.flags = GLib.OptionFlags.NONE
        option.long_name = GLib.OPTION_REMAINING
        option.short_name = 0
        return [option]

    def _startup(self) -> None:
        if sys.platform in ("win32", "darwin"):
            # Changing the PANGOCAIRO_BACKEND is necessary on Windows/MacOS
            # to render colored emoji glyphs
            os.environ["PANGOCAIRO_BACKEND"] = "fontconfig"

        app.ged.register_event_handler("db-migration", 0, self._on_db_migration)

        if not self._init_core():
            return

        icon_theme = get_icon_theme()
        icon_theme.add_search_path(str(configpaths.get("ICONS")))

        self.set_menubar(get_main_menu())

        from gajim.gtk import notification

        notification.init()

        self.avatar_storage = AvatarStorage()

        app.load_css_config()

        idle.Monitor.set_interval(
            app.settings.get("autoawaytime") * 60, app.settings.get("autoxatime") * 60
        )

        from gajim.gtk.status_icon import StatusIcon

        self.systray = StatusIcon()

        self._add_app_actions()
        accounts = app.settings.get_accounts()
        for account in accounts:
            self.add_account_actions(account)

        self._load_shortcuts()
        menus.build_accounts_menu()
        self.update_app_actions_state()

        self.register_event("feature-discovered", ged.CORE, self._on_feature_discovered)

        from gajim.gtk.main import MainWindow

        MainWindow()

        GLib.timeout_add(100, self._auto_connect)

    def _open_uris(self, uris: list[str]) -> None:
        accounts = app.settings.get_active_accounts()
        if not accounts:
            return

        for uri in uris:
            app.log("uri_handler").info("open %s", uri)
            if not uri.startswith("xmpp:"):
                continue

            try:
                iri, cmd = uri.split("?")
            except ValueError:
                # No query argument
                iri, cmd = uri, "message"

            try:
                jid = JID.from_iri(iri)
            except InvalidJid as error:
                app.log("uri_handler").warning("Invalid JID %s: %s", uri, error)
                continue

            if cmd == "join" and jid.resource:
                app.log("uri_handler").warning("Invalid MUC JID %s", uri)
                continue

            jid = str(jid)

            if cmd == "join":
                if len(accounts) == 1:
                    self.activate_action(
                        "open-chat", GLib.Variant("as", [accounts[0], jid])
                    )
                else:
                    self.activate_action("start-chat", GLib.Variant("as", [jid, ""]))

            elif cmd == "roster":
                self.activate_action("add-contact", GLib.Variant("s", jid))

            elif cmd.startswith("message"):
                attributes = cmd.split(";")
                message = None
                for key in attributes:
                    if not key.startswith("body"):
                        continue
                    try:
                        message = unquote(key.split("=")[1])
                    except Exception:
                        app.log("uri_handler").error("Invalid URI: %s", cmd)

                app.window.start_chat_from_jid(accounts[0], jid, message)

    def _shutdown(self, _application: GajimApplication) -> None:
        self._shutdown_core()

    def _quit_app(self) -> None:
        self.quit()

    def _command_line(
        self, _application: GajimApplication, command_line: Gio.ApplicationCommandLine
    ) -> int:
        """Handles command line options not related to the startup of Gajim."""

        options = command_line.get_options_dict()

        remote_commands = [
            ("start-chat", GLib.Variant("as", ["", ""])),
        ]

        for cmd, parameter in remote_commands:
            if options.contains(cmd):
                self.activate_action(cmd, parameter)
                return 0

        remaining = options.lookup_value(
            GLib.OPTION_REMAINING, GLib.VariantType.new("as")
        )

        if remaining is not None:
            self._open_uris(remaining.unpack())
            return 0

        if not options.contains("is-first-startup"):
            # If no commands have been handled and it's not the first
            # startup, raise the application.
            self.activate()

        return 0

    def _handle_local_options(
        self, _application: Gtk.Application, options: GLib.VariantDict
    ) -> int:

        if options.contains("version"):
            print(gajim.__version__)
            return 0

        profile = options.lookup_value("profile")
        if profile is None:
            GLib.set_application_name("Gajim")
        else:
            # Incorporate profile name into application id
            # to have a single app instance for each profile.
            profile = profile.get_string()
            GLib.set_application_name(f"Gajim ({profile})")
            app_id = f"{self.get_application_id()}.{profile}"
            self.set_application_id(app_id)
            configpaths.set_profile(profile)

        self.register()
        if self.get_is_remote():
            print(
                "Gajim is already running. "
                "The primary instance will handle remote commands"
            )
            return -1

        options.insert_value("is-first-startup", GLib.Variant("b", True))
        self._core_command_line(options)
        self._startup()
        return -1

    def _on_activate(self, _application: Gtk.Application) -> None:
        app.window.show_window()

    def _add_app_actions(self) -> None:
        for action in APP_ACTIONS:
            action_name, variant = action
            if variant is not None:
                variant = GLib.VariantType.new(variant)

            act = Gio.SimpleAction.new(action_name, variant)
            self.add_action(act)

        self._connect_app_actions()

    def _connect_app_actions(self) -> None:
        actions: ActionListT = [
            ("quit", self._on_quit_action),
            ("add-account", self._on_add_account_action),
            ("manage-proxies", self._on_manage_proxies_action),
            ("preferences", self._on_preferences_action),
            ("plugins", self._on_plugins_action),
            ("xml-console", self._on_xml_console_action),
            ("shortcuts", self._on_shortcuts_action),
            ("features", self._on_features_action),
            ("content", self._on_content_action),
            ("join-support-chat", self._on_join_support_chat),
            ("about", self._on_about_action),
            ("faq", self._on_faq_action),
            ("privacy-policy", self._on_privacy_policy_action),
            ("start-chat", self._on_new_chat_action),
            ("accounts", self._on_accounts_action),
            ("add-contact", self._on_add_contact_action),
            ("copy-text", self._on_copy_text_action),
            ("open-link", self._on_open_link_action),
            ("export-history", self._on_export_history_action),
            ("remove-history", self._on_remove_history_action),
            ("create-groupchat", self._on_create_groupchat_action),
            ("forget-groupchat", self._on_forget_groupchat_action),
            ("open-chat", self._on_open_chat_action),
            ("mute-chat", self._on_mute_chat_action),
        ]

        for action in actions:
            action_name, func = action
            act = self.lookup_action(action_name)
            assert act is not None
            act.connect("activate", func)
            self.add_action(act)

    def add_account_actions(self, account: str) -> None:
        for action_name, type_ in ACCOUNT_ACTIONS:
            account_action_name = f"{account}-{action_name}"
            if self.has_action(account_action_name):
                raise ValueError("Trying to add action more than once")

            act = Gio.SimpleAction.new(account_action_name, GLib.VariantType.new(type_))
            act.set_enabled(action_name in ALWAYS_ACCOUNT_ACTIONS)
            self.add_action(act)

        self._connect_account_actions(account)

    def _connect_account_actions(self, account: str) -> None:
        actions = [
            ("add-contact", self._on_add_contact_account_action),
            ("services", self._on_services_action),
            ("profile", self._on_profile_action),
            ("server-info", self._on_server_info_action),
            ("archive", self._on_archive_action),
            ("pep-config", self._on_pep_config_action),
            ("sync-history", self._on_sync_history_action),
            ("blocking", self._on_blocking_action),
            ("open-event", self._on_open_event_action),
            ("mark-as-read", self._on_mark_as_read_action),
            ("manage-roster", self._on_manage_roster_action),
            ("block-contact", self._on_block_contact),
            ("remove-contact", self._on_remove_contact),
            ("execute-command", self._on_execute_command),
        ]

        for action_name, func in actions:
            account_action_name = f"{account}-{action_name}"
            act = self.lookup_action(account_action_name)
            assert act is not None
            act.connect("activate", func)

    def remove_account_actions(self, account: str) -> None:
        for action_name in self.list_actions():
            if action_name.startswith(f"{account}-"):
                self.remove_action(action_name)

    def set_action_state(self, action_name: str, state: bool) -> None:
        action = self.lookup_action(action_name)
        assert isinstance(action, Gio.SimpleAction)
        action.set_enabled(state)

    def set_account_actions_state(self, account: str, new_state: bool = False) -> None:

        for action_name in ONLINE_ACCOUNT_ACTIONS:
            self.set_action_state(f"{account}-{action_name}", new_state)

        # Disable all feature actions on disconnect
        if not new_state:
            for action_name in FEATURE_ACCOUNT_ACTIONS:
                self.set_action_state(f"{account}-{action_name}", new_state)

    def update_feature_actions_state(self, account: str) -> None:
        client = app.get_client(account)
        mam_available = client.get_module("MAM").available
        blocking_available = client.get_module("Blocking").supported

        self.set_action_state(f"{account}-archive", mam_available)
        self.set_action_state(f"{account}-blocking", blocking_available)
        self.set_action_state(f"{account}-block-contact", blocking_available)

    def update_app_actions_state(self) -> None:
        active_accounts = bool(app.get_connected_accounts(exclude_local=True))
        self.set_action_state("create-groupchat", active_accounts)

        enabled_accounts = bool(app.settings.get_active_accounts())
        self.set_action_state("start-chat", enabled_accounts)

    def _load_shortcuts(self) -> None:
        default_path = configpaths.get("DATA") / "other" / "shortcuts.json"
        shortcuts = load_json(default_path)
        assert shortcuts is not None

        user_path = configpaths.get("MY_SHORTCUTS")
        user_shortcuts = {}
        if user_path.exists():
            app.log("app").info("Load user shortcuts")
            user_shortcuts = load_json(user_path, default={})

        shortcuts.update(user_shortcuts)

        for action, accels in shortcuts.items():
            self.set_accels_for_action(action, accels)

    def _on_feature_discovered(self, event: events.FeatureDiscovered) -> None:
        self.update_feature_actions_state(event.account)

    def create_account(
        self,
        account: str,
        address: JID,
        password: str,
        proxy_name: str | None,
        custom_host: tuple[str, ConnectionProtocol, ConnectionType] | None,
        anonymous: bool = False,
    ) -> None:

        CoreApplication.create_account(
            self, account, address, password, proxy_name, custom_host, anonymous
        )

        app.css_config.refresh()

        # Action must be added before account window is updated
        self.add_account_actions(account)

        window = get_app_window("AccountsWindow")
        if window is not None:
            window.add_account(account)

    def enable_account(self, account: str) -> None:
        CoreApplication.enable_account(self, account)
        menus.build_accounts_menu()
        self.update_app_actions_state()
        window = get_app_window("AccountsWindow")
        if window is not None:
            window.enable_account(account, True)

    def disable_account(self, account: str) -> None:
        for win in get_app_windows(account):
            # Close all account specific windows, except the RemoveAccount
            # dialog. It shows if the removal was successful.
            if type(win).__name__ == "RemoveAccount":
                continue
            win.close()

        CoreApplication.disable_account(self, account)

        menus.build_accounts_menu()
        self.update_app_actions_state()

    def remove_account(self, account: str) -> None:
        CoreApplication.remove_account(self, account)

        self.remove_account_actions(account)

        window = get_app_window("AccountsWindow")
        if window is not None:
            window.remove_account(account)

    def _on_db_migration(self, _event: events.DBMigration) -> None:
        open_window("DBMigration")

        context = GLib.MainContext.default()
        while context.pending():
            context.iteration(may_block=False)

    # Action Callbacks

    @staticmethod
    def _on_add_contact_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        jid = param.get_string() or None
        if jid is not None:
            jid = JID.from_string(jid)
        open_window("AddContact", account=None, jid=jid)

    @staticmethod
    def _on_preferences_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Preferences")

    @staticmethod
    def _on_plugins_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("PluginsWindow")

    @staticmethod
    def _on_accounts_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        window = open_window("AccountsWindow")

        account = param.get_string()
        if account:
            window.select_account(account)

    @staticmethod
    def _on_quit_action(_action: Gio.SimpleAction, _param: GLib.Variant | None) -> None:
        app.window.quit()

    @staticmethod
    def _on_new_chat_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:

        jid, initial_message = param.get_strv()
        open_window(
            "StartChatDialog",
            initial_jid=jid or None,
            initial_message=initial_message or None,
        )

    @staticmethod
    def _on_profile_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("ProfileWindow", account=account)

    @staticmethod
    def _on_services_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("ServiceDiscoveryWindow", account=account, address_entry=True)

    @staticmethod
    def _on_create_groupchat_action(
        _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account = param.get_string()
        open_window("CreateGroupchatWindow", account=account or None)

    @structs.actionmethod
    def _on_add_contact_account_action(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        open_window("AddContact", account=params.account, jid=params.jid)

    @staticmethod
    def _on_add_account_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("AccountWizard")

    @staticmethod
    def _on_manage_roster_action(
        _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account = param.get_string()
        open_window("ManageRoster", account=account)

    @structs.actionmethod
    def _on_block_contact(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        app.window.block_contact(params.account, params.jid)

    def _on_remove_contact(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account, jid = param.unpack()
        app.window.remove_contact(account, JID.from_string(jid))

    def _on_execute_command(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        account, jids = param.unpack()
        open_window("AdHocCommands", account=account, jids=jids)

    @staticmethod
    def _on_pep_config_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("PEPConfig", account=account)

    @staticmethod
    def _on_archive_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("MamPreferences", account=account)

    @staticmethod
    def _on_blocking_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("BlockingList", account=account)

    @staticmethod
    def _on_sync_history_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("HistorySyncAssistant", account=account)

    @staticmethod
    def _on_server_info_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account = param.get_string()
        open_window("ServerInfo", account=account)

    @staticmethod
    def _on_xml_console_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("DebugConsoleWindow")

    @staticmethod
    def _on_manage_proxies_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("ManageProxies")

    @staticmethod
    def _on_content_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_uri(GAJIM_WIKI_URI)

    @staticmethod
    def _on_join_support_chat(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        accounts = app.settings.get_active_accounts()
        if len(accounts) == 1:
            app.window.show_add_join_groupchat(accounts[0], GAJIM_SUPPORT_JID)
            return
        open_window("StartChatDialog", initial_jid=GAJIM_SUPPORT_JID)

    @staticmethod
    def _on_faq_action(_action: Gio.SimpleAction, _param: GLib.Variant | None) -> None:
        open_uri(GAJIM_FAQ_URI)

    @staticmethod
    def _on_privacy_policy_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_uri(GAJIM_PRIVACY_POLICY_URI)

    @staticmethod
    def _on_shortcuts_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        ShortcutsWindow()

    @staticmethod
    def _on_features_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        open_window("Features")

    @staticmethod
    def _on_about_action(
        _action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> None:
        AboutDialog()

    # @staticmethod
    # TODO Jingle FT
    # def _on_file_transfer_action(_action: Gio.SimpleAction,
    #                              _param: GLib.Variant | None) -> None:

    #     ft = app.interface.instances['file_transfers']
    #     if ft.window.get_property('visible'):
    #         ft.window.present()
    #     else:
    #         ft.window.show_all()

    @staticmethod
    @structs.actionfunction
    def _on_open_event_action(
        _action: Gio.SimpleAction, params: structs.OpenEventActionParams
    ) -> None:

        if params.type in (
            "connection-failed",
            "subscription-request",
            "unsubscribed",
            "group-chat-invitation",
            "server-shutdown",
        ):

            app.window.show_account_page(params.account)

        elif params.type in ("incoming-message", "incoming-call"):

            assert params.jid
            jid = JID.from_string(params.jid)
            app.window.select_chat(params.account, jid)

        app.window.present()

    @structs.actionmethod
    def _on_mark_as_read_action(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:

        app.window.mark_as_read(params.account, params.jid)

    @staticmethod
    def _on_open_link_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account, uri = param.get_strv()
        open_uri(uri, account=account)

    @staticmethod
    def _on_copy_text_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        app.window.get_clipboard().set(param.get_string())

    @staticmethod
    def _on_open_chat_action(_action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account, jid = param.get_strv()
        app.window.start_chat_from_jid(account, jid)

    @staticmethod
    @structs.actionfunction
    def _on_mute_chat_action(
        _action: Gio.SimpleAction, params: structs.MuteContactParam
    ) -> None:

        client = app.get_client(params.account)
        contact = client.get_module("Contacts").get_contact(params.jid)
        assert not isinstance(contact, ResourceContact)

        if params.state == MuteState.UNMUTE:
            contact.settings.set("mute_until", None)
            return

        until = datetime.now(UTC) + timedelta(minutes=params.state)
        contact.settings.set("mute_until", until.isoformat())

    @staticmethod
    @structs.actionfunction
    def _on_export_history_action(
        _action: Gio.SimpleAction, params: structs.ExportHistoryParam
    ) -> None:
        open_window("HistoryExport", account=params.account, jid=params.jid)

    @staticmethod
    @structs.actionfunction
    def _on_remove_history_action(
        _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        def _remove() -> None:
            app.storage.archive.remove_history_for_jid(params.account, params.jid)

            app.window.clear_chat_list_row(params.account, params.jid)
            control = app.window.get_control()
            if not control.is_loaded(params.account, params.jid):
                return

            control.reset_view()

        ConfirmationDialog(
            _("Remove Chat History?"),
            _("Do you want to remove all chat history for this chat?"),
            [
                DialogButton.make("Cancel"),
                DialogButton.make("Remove", callback=_remove),
            ],
            transient_for=app.window,
        )

    @staticmethod
    @structs.actionfunction
    def _on_forget_groupchat_action(
        _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:

        def _forget():
            window = get_app_window("StartChatDialog")
            if window is not None:
                window.remove_row(params.account, str(params.jid))

            client = app.get_client(params.account)
            client.get_module("MUC").leave(params.jid)
            client.get_module("Bookmarks").remove(params.jid)

            app.storage.archive.remove_history_for_jid(params.account, params.jid)

        ConfirmationDialog(
            _("Forget this Group Chat?"),
            _("Do you want to remove this chat including its chat history?"),
            [
                DialogButton.make("Cancel"),
                DialogButton.make("Remove", callback=_forget),
            ],
            transient_for=app.window,
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import cairo
import nbxmpp
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.const import ARTISTS
from gajim.common.const import DEVELOPERS
from gajim.common.const import MAINTAINERS
from gajim.common.const import THANKS
from gajim.common.i18n import _
from gajim.common.util.app import get_extended_app_version
from gajim.common.util.uri import open_uri
from gajim.common.util.version import get_glib_version
from gajim.common.util.version import get_gobject_version
from gajim.common.util.version import get_soup_version

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_gtk_version


class AboutDialog(Gtk.AboutDialog, SignalManager):
    def __init__(self):
        Gtk.AboutDialog.__init__(
            self,
            transient_for=app.window,
            name="Gajim",
            version=get_extended_app_version(),
            copyright="Copyright © 2003-2025 Gajim Team",
            license_type=Gtk.License.GPL_3_0_ONLY,
            website="https://gajim.org/",
            logo_icon_name="gajim",
            translator_credits=_("translator-credits"),
        )
        SignalManager.__init__(self)

        cairo_ver = cairo.cairo_version_string()
        python_cairo_ver = cairo.version

        comments: list[str] = []
        comments.append(_("A fully-featured XMPP chat client"))
        comments.append("")
        comments.append(_("GTK Version: %s") % get_gtk_version())
        comments.append(_("GLib Version: %s") % get_glib_version())
        comments.append(_("Pango Version: %s") % Pango.version_string())
        comments.append(_("PyGObject Version: %s") % get_gobject_version())
        comments.append(_("cairo Version: %s") % cairo_ver)
        comments.append(_("pycairo Version: %s") % python_cairo_ver)
        comments.append(_("python-nbxmpp Version: %s") % nbxmpp.__version__)
        comments.append(_("libsoup Version: %s") % get_soup_version())

        self.set_comments("\n".join(comments))

        self.add_credit_section(_("Maintainers"), MAINTAINERS)
        self.add_credit_section(_("Developers"), DEVELOPERS)
        self.add_credit_section(_("Artists"), ARTISTS)

        thanks = list(THANKS)
        thanks.append("")
        thanks.append(_("Last but not least"))
        thanks.append(_("we would like to thank all the package maintainers."))
        self.add_credit_section(_("Thanks"), thanks)

        self._connect(self, "activate-link", self._on_activate_link)
        self._connect(self, "close-request", self._on_close_request)
        self.set_visible(True)

    def _on_close_request(self, window: AboutDialog) -> None:
        self._disconnect_all()
        app.check_finalize(self)

    @staticmethod
    def _on_activate_link(_label: Gtk.Label, uri: str) -> int:
        # We have to use this, because the default GTK handler
        # is not cross-platform compatible
        open_uri(uri)
        return Gdk.EVENT_STOP



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import cast

import functools
import importlib.metadata
import os
import platform
import sys

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Soup
from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet

CURRENT_PYTHON_VERSION = platform.python_version()


def python_version(specifier_set: str) -> bool:
    spec = SpecifierSet(specifier_set)
    return CURRENT_PYTHON_VERSION in spec


def package_version(requirement: str) -> bool:
    req = Requirement(requirement)

    try:
        installed_version = importlib.metadata.version(req.name)
    except importlib.metadata.PackageNotFoundError:
        return False

    return installed_version in req.specifier


@functools.lru_cache(maxsize=1)
def get_os_info() -> str:
    info = 'N/A'
    if sys.platform in ('win32', 'darwin'):
        info = f'{platform.system()} {platform.release()}'

    elif sys.platform == 'linux':
        try:
            import distro  # type: ignore

            info = cast(str, distro.name(pretty=True))  # type: ignore
        except ImportError:
            info = platform.system()
    return info


def get_os_name() -> str:
    if sys.platform in ('win32', 'darwin'):
        return platform.system()
    if os.name == 'posix':
        try:
            import distro  # type: ignore

            return distro.name(pretty=True)  # type: ignore
        except ImportError:
            return platform.system()
    return ''


def get_os_version() -> str:
    if sys.platform in ('win32', 'darwin'):
        return platform.version()
    if os.name == 'posix':
        try:
            import distro  # type: ignore

            return distro.version(pretty=True)  # type: ignore
        except ImportError:
            return platform.release()
    return ''


def get_gobject_version() -> str:
    return '.'.join(map(str, GObject.pygobject_version))


def get_glib_version() -> str:
    return '.'.join(
        map(str, [GLib.MAJOR_VERSION, GLib.MINOR_VERSION, GLib.MICRO_VERSION])
    )


def get_soup_version() -> str:
    return '.'.join(
        map(
            str,
            [
                Soup.get_major_version(),
                Soup.get_minor_version(),
                Soup.get_micro_version(),
            ],
        )
    )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import cast

import logging
import math
from collections.abc import Callable
from pathlib import Path

try:
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst

from gajim.common import app
from gajim.common.preview import AudioSampleT

log = logging.getLogger("gajim.gtk.preview_audio_analyzer")


class AudioAnalyzer:
    def __init__(
        self,
        filepath: Path,
        duration_callback: Callable[[float], None],
        samples_callback: Callable[[AudioSampleT], None],
    ) -> None:

        if not app.is_installed("GST"):
            log.info(
                "Could not create AudioAnalyzer because GStreamer is not installed."
            )
            return

        self._playbin = Gst.ElementFactory.make("playbin")

        if self._playbin is None:
            log.error("Could not create GST playbin for AudioAnalyzer")
            return

        self._duration_callback = duration_callback
        self._duration_updated = False
        self._samples_callback = samples_callback
        self._query = Gst.Query.new_position(Gst.Format.TIME)
        self._duration = Gst.CLOCK_TIME_NONE  # in ns
        self._num_channels = 1
        self._samples: list[tuple[float, float]] = []
        self._level: Gst.Element | None = None
        self._bus_watch_id: int = 0

        if filepath.is_file():
            self._setup_audio_analyzer(filepath)

    def _setup_audio_analyzer(self, file_path: Path) -> None:
        assert isinstance(self._playbin, Gst.Bin)

        audio_sink = Gst.Bin.new("audiosink")
        audioconvert = Gst.ElementFactory.make("audioconvert", "audioconvert")
        self._level = Gst.ElementFactory.make("level", "level")
        fakesink = Gst.ElementFactory.make("fakesink", "fakesink")

        pipeline_elements = [
            audio_sink,
            audioconvert,
            self._level,
            fakesink,
        ]

        if any(element is None for element in pipeline_elements):
            log.error("Could not set up pipeline for AudioAnalyzer")
            return

        assert audioconvert is not None
        assert self._level is not None
        assert fakesink is not None

        audio_sink.add(audioconvert)
        audio_sink.add(self._level)
        audio_sink.add(fakesink)

        audioconvert.link(self._level)
        self._level.link(fakesink)

        sink_pad = audioconvert.get_static_pad("sink")
        assert sink_pad is not None
        ghost_pad = Gst.GhostPad.new("sink", sink_pad)
        assert ghost_pad is not None
        audio_sink.add_pad(ghost_pad)

        self._playbin.set_property("audio-sink", audio_sink)
        if file_path.is_file():
            file_uri = file_path.as_uri()
            self._playbin.set_property("uri", file_uri)
        self._playbin.no_more_pads()

        # TODO: Message is no longer a property: Check if required
        # self._level.set_property('message', True)
        fakesink.set_property("sync", False)

        state_return = self._playbin.set_state(Gst.State.PLAYING)
        if state_return == Gst.StateChangeReturn.FAILURE:
            log.warning("Could not set up GST playbin")
            return

        self._level_element = self._playbin.get_by_name("level")
        bus = self._playbin.get_bus()
        if bus is None:
            log.error("Could not get GST Bus")
            return

        bus.add_signal_watch()
        self._bus_watch_id = bus.connect("message", self._on_bus_message)

    def _on_bus_message(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        assert self._playbin is not None

        if message.type == Gst.MessageType.EOS:
            self._samples_callback(self._samples)
            self._playbin.set_state(Gst.State.NULL)
            return

        if message.type in (
            Gst.MessageType.STATE_CHANGED,
            Gst.MessageType.DURATION_CHANGED,
        ):
            _success, self._duration = self._playbin.query_duration(Gst.Format.TIME)
            if not self._duration_updated:
                if _success:
                    assert self._duration is not None
                    self._duration_callback(float(self._duration))
                    self._duration_updated = True
            return

        if message.src is self._level:
            structure = message.get_structure()
            if (
                structure is None
                or structure.get_name() != "level"
                or not structure.has_field("rms")
            ):
                return

            # RMS: Root Mean Square = Average Power
            rms_values = cast(list[float], structure.get_value("rms"))
            assert rms_values is not None
            self._num_channels = min(2, len(rms_values))

            # Convert from dB to a linear scale.
            # The sound pressure level L is defined as
            # L = 10 log_10((p/p_0)^2) dB, where p is the RMS value
            # of the sound pressure.
            if self._num_channels == 0:
                self._samples.append((0, 0))
            elif self._num_channels == 1:
                lin_val = math.pow(10, rms_values[0] / 10 / 2)
                self._samples.append((lin_val, lin_val))
            else:
                lin_val1 = math.pow(10, rms_values[0] / 10 / 2)
                lin_val2 = math.pow(10, rms_values[1] / 10 / 2)
                self._samples.append((lin_val1, lin_val2))

    def destroy(self) -> None:
        if self._playbin is not None:
            self._playbin.set_state(Gst.State.NULL)
            bus = self._playbin.get_bus()

            if bus is not None:
                bus.remove_signal_watch()
                bus.disconnect(self._bus_watch_id)

        del self._duration_callback, self._samples_callback
        app.check_finalize(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.events import MucInvitation
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.util.muc import get_group_chat_nick

from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.groupchat_nick import NickChooser
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.widgets import AccountBadge
from gajim.gtk.widgets import GajimAppWindow


class GroupChatInvitationDialog(GajimAppWindow):
    def __init__(self, account: str, event: MucInvitation) -> None:
        GajimAppWindow.__init__(
            self,
            name="GroupChatInvitationDialog",
            title=_("Group Chat Invitation"),
            default_width=450,
            default_height=500,
        )

        self.account = account

        invitation_widget = GroupChatInvitation(account, event)
        self._connect(invitation_widget, "accepted", self._on_invitation_widget_action)
        self._connect(invitation_widget, "declined", self._on_invitation_widget_action)
        self.set_child(invitation_widget)

    def _on_invitation_widget_action(self, _widget: GroupChatInvitation) -> None:
        self.close()

    def _cleanup(self) -> None:
        pass


class GroupChatInvitation(Gtk.Box, SignalManager):

    __gsignals__ = {
        "accepted": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        ),
        "declined": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        ),
    }

    def __init__(self, account: str, event: MucInvitation) -> None:
        Gtk.Box.__init__(self, halign=Gtk.Align.CENTER)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self._room_jid = str(event.muc)
        self._from = str(event.from_)
        self._password = event.password

        main_box = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=12, valign=Gtk.Align.FILL
        )

        muc_info_box = GroupChatInfoScrolled(account, minimal=True)
        muc_info_box.set_from_disco_info(event.info)
        main_box.append(muc_info_box)

        separator = Gtk.Separator()
        main_box.append(separator)

        contact = self._client.get_module("Contacts").get_contact(event.from_.bare)
        assert isinstance(contact, BareContact | GroupchatContact)
        contact_label = Gtk.Label(label=contact.name, wrap=True)
        contact_label.add_css_class("bold16")
        contact_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, spacing=6, halign=Gtk.Align.CENTER
        )
        contact_box.append(contact_label)
        main_box.append(contact_box)

        enabled_accounts = app.get_enabled_accounts_with_labels()
        if len(enabled_accounts) > 1:
            account_badge = AccountBadge(account)
            contact_box.append(account_badge)

        invitation_label = Gtk.Label(
            label=_("has invited you to a group chat.\nDo you want to join?"),
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
            max_width_chars=50,
            wrap=True,
        )
        main_box.append(invitation_label)

        decline_button = Gtk.Button.new_with_mnemonic(_("_Decline"))
        decline_button.set_halign(Gtk.Align.START)
        decline_button.set_margin_end(24)
        self._connect(decline_button, "clicked", self._on_decline)

        self._nick_chooser = NickChooser()
        self._nick_chooser.set_text(get_group_chat_nick(self._account, event.info.jid))

        join_button = Gtk.Button.new_with_mnemonic(_("_Join"))
        join_button.set_halign(Gtk.Align.END)
        join_button.add_css_class("suggested-action")
        self._connect(join_button, "clicked", self._on_join)

        join_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, halign=Gtk.Align.END, hexpand=True
        )
        join_box.add_css_class("linked")
        join_box.append(self._nick_chooser)
        join_box.append(join_button)

        button_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL,
            margin_top=6,
            vexpand=True,
            valign=Gtk.Align.END,
        )
        button_box.prepend(decline_button)
        button_box.append(join_box)

        main_box.append(button_box)
        self.append(main_box)

        join_button.grab_focus()

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def _on_join(self, _button: Gtk.Button) -> None:
        nickname = self._nick_chooser.get_text()
        app.window.show_add_join_groupchat(
            self._account, self._room_jid, nickname=nickname
        )
        self.emit("accepted")

    def _on_decline(self, _button: Gtk.Button) -> None:
        self._client.get_module("MUC").decline(self._room_jid, self._from)
        self.emit("declined")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import inspect
import logging
from importlib import import_module
from pathlib import Path

from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common.modules.base import BaseModule

if typing.TYPE_CHECKING:
    from gajim.common.types import Client


log = logging.getLogger('gajim.c.m')

_modules: dict[str, dict[str, BaseModule]] = {}
_store_publish_modules = [
    'UserLocation',
    'UserTune',
]


def register_modules(client: Client) -> None:
    if client.account in _modules:
        return

    _modules[client.account] = {}

    path = Path(__file__).parent
    for module in path.iterdir():
        if module.suffix not in ['.py', '.pyc']:
            continue

        if module.name.startswith('__'):
            continue

        name = module.stem
        module = import_module('.%s' % name, package='gajim.common.modules')
        for _, base_class in inspect.getmembers(module, inspect.isclass):

            if base_class is BaseModule:
                continue

            if BaseModule not in inspect.getmro(base_class):
                continue

            instance = base_class.get_instance(client)
            module_name = base_class.__name__
            _modules[client.account][module_name] = instance


def register_single_module(client: Client,
                           instance: BaseModule,
                           name: str) -> None:

    if client.account not in _modules:
        raise ValueError('Unknown account name: %s' % client.account)
    _modules[client.account][name] = instance


def unregister_modules(client: Client) -> None:
    for instance in _modules[client.account].values():
        if hasattr(instance, 'cleanup'):
            instance.cleanup()
        app.check_finalize(instance)
    del _modules[client.account]


def unregister_single_module(client: Client, name: str) -> None:
    if client.account not in _modules:
        return
    if name not in _modules[client.account]:
        return
    del _modules[client.account][name]


def send_stored_publish(account: str) -> None:
    for name in _store_publish_modules:
        _modules[account][name].send_stored_publish()


def get(account: str, name: str) -> BaseModule:
    return _modules[account][name]


def get_handlers(client: Client) -> list[StanzaHandler]:
    handlers: list[StanzaHandler] = []
    for module in _modules[client.account].values():
        handlers += module.handlers
    return handlers



# Copyright (C) 2005 Alex Mauer <hawke AT hawkesnest.net>
#                    Stéphan Kochen <stephan AT kochen.nl>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006 Travis Shirk <travis AT pobox.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2007 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2007-2008 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
from datetime import datetime

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.util.status import get_uf_show
from gajim.common.util.text import format_location
from gajim.common.util.text import format_tune
from gajim.common.util.user_strings import get_uf_affiliation
from gajim.common.util.user_strings import get_uf_sub

from gajim.gtk.avatar import get_show_circle
from gajim.gtk.builder import get_builder
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import convert_surface_to_texture
from gajim.gtk.util.misc import iterate_children

log = logging.getLogger("gajim.gtk.tooltips")


class GCTooltip:
    def __init__(self) -> None:
        self._contact: GroupchatParticipant | None = None

    def clear_tooltip(self) -> None:
        self._contact = None

    def get_tooltip(self, contact: GroupchatParticipant) -> tuple[bool, Gtk.Grid]:
        if not hasattr(self, "_ui"):
            self._ui = get_builder("groupchat_roster_tooltip.ui")

        if self._contact == contact:
            return True, self._ui.tooltip_grid

        self._populate_grid(contact)
        self._contact = contact
        return False, self._ui.tooltip_grid

    def _hide_grid_children(self) -> None:
        """
        Hide all Elements of the Tooltip Grid
        """
        for widget in iterate_children(self._ui.tooltip_grid):
            widget.set_visible(False)

    def _populate_grid(self, contact: GroupchatParticipant) -> None:
        """
        Populate the Tooltip Grid with data of from the contact
        """
        self._hide_grid_children()

        self._ui.nick.set_text(contact.name)
        self._ui.nick.set_visible(True)

        # Status Message
        if contact.status:
            status = contact.status.strip()
            if status != "":
                self._ui.status.set_text(status)
                self._ui.status.set_visible(True)

        # JID
        if contact.real_jid is not None:
            self._ui.jid.set_text(str(contact.real_jid.bare))
            self._ui.jid.set_visible(True)

        # Affiliation
        if not contact.affiliation.is_none:
            uf_affiliation = get_uf_affiliation(contact.affiliation)
            uf_affiliation = _("%(owner_or_admin_or_member)s of this group chat") % {
                "owner_or_admin_or_member": uf_affiliation
            }
            self._ui.affiliation.set_text(uf_affiliation)
            self._ui.affiliation.set_visible(True)

        if contact.hats is not None:

            container_remove_all(self._ui.hats_box)

            for hat in contact.hats.get_hats()[:5]:
                # Limit to 5 hats
                hat_badge = Gtk.Box(spacing=6, halign=Gtk.Align.START)
                hat_badge.add_css_class("badge")
                hat_badge.add_css_class("hat-badge")

                hat_badge_icon = Gtk.Image.new_from_icon_name("feather-tag-symbolic")
                hat_badge.append(hat_badge_icon)

                hat_badge_label = Gtk.Label(
                    label=GLib.markup_escape_text(hat.title),
                    ellipsize=Pango.EllipsizeMode.END,
                    max_width_chars=20,
                    halign=Gtk.Align.START,
                )
                hat_badge.append(hat_badge_label)

                self._ui.hats_box.append(hat_badge)
                self._ui.hats_box.set_visible(True)

        # Avatar
        scale = self._ui.tooltip_grid.get_scale_factor()
        texture = contact.get_avatar(AvatarSize.TOOLTIP, scale)
        self._ui.avatar.set_pixel_size(AvatarSize.TOOLTIP)
        self._ui.avatar.set_from_paintable(texture)
        self._ui.avatar.set_visible(True)
        self._ui.fillelement.set_visible(True)

        app.plugin_manager.extension_point(
            "gc_tooltip_populate", self, contact, self._ui.tooltip_grid
        )


class ContactTooltip:
    def __init__(self) -> None:
        self._contact = None

    def clear_tooltip(self) -> None:
        self._contact = None
        if not hasattr(self, "_ui"):
            return
        container_remove_all(self._ui.resources_box)
        for widget in iterate_children(self._ui.tooltip_grid):
            widget.set_visible(False)

    def get_tooltip(self, contact: types.BareContact) -> tuple[bool, Gtk.Grid]:
        if not hasattr(self, "_ui"):
            self._ui = get_builder("contact_tooltip.ui")

        if self._contact == contact:
            return True, self._ui.tooltip_grid

        self.clear_tooltip()
        self._populate_grid(contact)
        self._contact = contact
        return False, self._ui.tooltip_grid

    def _populate_grid(self, contact: types.BareContact) -> None:
        scale = self._ui.tooltip_grid.get_scale_factor()

        texture = contact.get_avatar(AvatarSize.TOOLTIP, scale)
        self._ui.avatar.set_pixel_size(AvatarSize.TOOLTIP)
        self._ui.avatar.set_from_paintable(texture)
        self._ui.avatar.set_visible(True)

        self._ui.name.set_markup(GLib.markup_escape_text(contact.name))
        self._ui.name.set_visible(True)

        self._ui.jid.set_text(str(contact.jid))
        self._ui.jid.set_visible(True)

        if contact.has_resources():
            for res in contact.iter_resources():
                self._add_resources(res)
        else:
            self._add_resources(contact)

        if contact.subscription and contact.subscription != "both":
            # 'both' is the normal subscription value, just omit it
            self._ui.sub.set_text(get_uf_sub(contact.subscription))
            self._ui.sub.set_visible(True)
            self._ui.sub_label.set_visible(True)

        self._append_pep_info(contact)

        app.plugin_manager.extension_point("contact_tooltip_populate", self, contact)

        # This sets the bottom-most widget to expand, in case the avatar
        # takes more space than the labels
        row_count = 1
        last_widget = self._ui.tooltip_grid.get_child_at(1, 1)
        assert last_widget is not None
        while row_count < 6:
            widget = self._ui.tooltip_grid.get_child_at(1, row_count)
            if widget and widget.get_visible():
                last_widget = widget
                row_count += 1
            else:
                break
        last_widget.set_vexpand(True)

    def _add_resources(
        self, contact: types.BareContact | types.ResourceContact
    ) -> None:

        resource_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        show_surface = get_show_circle(
            contact.show,
            AvatarSize.SHOW_CIRCLE,
            self._ui.tooltip_grid.get_scale_factor(),
        )
        show_image = Gtk.Image(
            paintable=convert_surface_to_texture(show_surface),
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
        )

        show_string = get_uf_show(contact.show.value)

        resource_string = ""
        if contact.jid.resource is not None:
            escaped_resource = GLib.markup_escape_text(contact.jid.resource)
            resource_string = f" ({escaped_resource})"

        resource_label = Gtk.Label(
            ellipsize=Pango.EllipsizeMode.END,
            halign=Gtk.Align.START,
            label=f"{show_string}{resource_string}",
            max_width_chars=30,
            xalign=0,
        )

        base_box = Gtk.Box(spacing=6)
        base_box.append(show_image)
        base_box.append(resource_label)
        resource_box.append(base_box)

        if contact.status:
            status_label = Gtk.Label(
                ellipsize=Pango.EllipsizeMode.END,
                halign=Gtk.Align.START,
                label=contact.status,
                max_width_chars=30,
                xalign=0,
            )
            resource_box.append(status_label)

        if idle_datetime := contact.idle_datetime:
            current = datetime.now()
            if idle_datetime.date() == current.date():
                format_string = app.settings.get("time_format")
                formatted = idle_datetime.strftime(format_string)
            else:
                format_string = app.settings.get("date_time_format")
                formatted = idle_datetime.strftime(format_string)
            idle_text = _("Idle since: %s") % formatted
            idle_label = Gtk.Label(
                halign=Gtk.Align.START,
                label=idle_text,
                xalign=0,
            )
            resource_box.append(idle_label)

        app.plugin_manager.extension_point(
            "contact_tooltip_resource_populate", resource_box, contact
        )

        self._ui.resources_box.append(resource_box)
        self._ui.resources_box.set_visible(True)

    def _append_pep_info(self, contact: types.BareContact) -> None:
        tune = contact.get_tune()
        if tune is not None:
            tune_str = format_tune(tune)
            self._ui.tune.set_markup(tune_str)
            self._ui.tune.set_visible(True)
            self._ui.tune_label.set_visible(True)

        location = contact.get_location()
        if location is not None:
            location_str = format_location(location)
            self._ui.location.set_markup(location_str)
            self._ui.location.set_visible(True)
            self._ui.location_label.set_visible(True)
